# -*- Mode: Python; coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2009 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Henrique Romano <henrique@async.com.br>
##              Johan Dahlin <henrique@async.com.br>
##
"""
FiscNet base driver implementation.
"""

import ConfigParser
import datetime
from decimal import Decimal
import logging
import re
import os

from kiwi.currency import currency
from kiwi.python import Settable
from serial import PARITY_EVEN
from zope.interface import implements

from stoqdrivers.enum import PaymentMethodType, TaxType, UnitType
from stoqdrivers.exceptions import (DriverError, PendingReduceZ,
                                    CommandParametersError, CommandError,
                                    ReadXError, OutofPaperError,
                                    CouponTotalizeError, PaymentAdditionError,
                                    CancelItemError, CouponOpenError,
                                    InvalidState, PendingReadX,
                                    CloseCouponError, CouponNotOpenError)
from stoqdrivers.interfaces import ICouponPrinter, IChequePrinter
from stoqdrivers.printers.capabilities import Capability
from stoqdrivers.printers.cheque import BaseChequePrinter, BankConfiguration
from stoqdrivers.printers.base import BaseDriverConstants
from stoqdrivers.translation import stoqdrivers_gettext
from stoqdrivers.serialbase import SerialBase

_ = stoqdrivers_gettext

log = logging.getLogger('stoqdrivers.conf')

# Page 92
[FLAG_INTERVENCAO_TECNICA,
 FLAG_SEM_MFD,
 FLAG_RAM_NOK,
 FLAG_RELOGIO_NOK,
 FLAG_SEM_MF,
 FLAG_DIA_FECHADO,
 FLAG_DIA_ABERTO,
 FLAG_Z_PENDENTE,
 FLAG_SEM_PAPEL,
 FLAG_MECANISM_NOK,
 FLAG_DOCUMENTO_ABERTO,
 FLAG_INSCRICOES_OK,
 FLAG_CLICHE_OK,
 FLAG_EM_LINHA,
 FLAG_MFD_ESGOTADA] = _status_flags = [2 ** n for n in range(15)]

_flagnames = {
    FLAG_INTERVENCAO_TECNICA: 'FLAG_INTERVENCAO_TECNICA',
    FLAG_SEM_MFD: 'FLAG_SEM_MFD',
    FLAG_RAM_NOK: 'FLAG_RAM_NOK',
    FLAG_RELOGIO_NOK: 'FLAG_RELOGIO_NOK',
    FLAG_SEM_MF: 'FLAG_SEM_MF',
    FLAG_DIA_FECHADO: 'FLAG_DIA_FECHADO',
    FLAG_DIA_ABERTO: 'FLAG_DIA_ABERTO',
    FLAG_Z_PENDENTE: 'FLAG_Z_PENDENTE',
    FLAG_SEM_PAPEL: 'FLAG_SEM_PAPEL',
    FLAG_MECANISM_NOK: 'FLAG_MECANISM_NOK',
    FLAG_DOCUMENTO_ABERTO: 'FLAG_DOCUMENTO_ABERTO',
    FLAG_INSCRICOES_OK: 'FLAG_INSCRICOES_OK',
    FLAG_CLICHE_OK: 'FLAG_CLICHE_OK',
    FLAG_EM_LINHA: 'FLAG_EM_LINHA',
    FLAG_MFD_ESGOTADA: 'FLAG_MFD_ESGOTADA',
}


class FiscNetConstants(BaseDriverConstants):
    _constants = {
        UnitType.WEIGHT: 'km',
        UnitType.LITERS: 'lt',
        UnitType.METERS: 'm ',
        UnitType.EMPTY: '  ',
        PaymentMethodType.MONEY: '-2',

        # FIXME: This should not be hardcoded (is it still)?
        #         PaymentMethodType.CHECK:        '2',
        #         PaymentMethodType.MONEY: '-2',
        #         PaymentMethodType.CHECK: '0',
        #         PaymentMethodType.BOLETO: '1',
        #         PaymentMethodType.CREDIT_CARD: '2',
        #         PaymentMethodType.DEBIT_CARD: '3',
        #         PaymentMethodType.FINANCIAL: '4',
        #         PaymentMethodType.GIFT_CERTIFICATE: '5,
    }

_RETVAL_TOKEN_RE = re.compile(r"^\s*([^=\s;]+)")
_RETVAL_QUOTED_VALUE_RE = re.compile(r"^\s*=\s*\"([^\"\\]*(?:\\.[^\"\\]*)*)\"")
_RETVAL_VALUE_RE = re.compile(r"^\s*=\s*([^\s;]*)")
_RETVAL_ESCAPE_RE = re.compile(r"\\(.)")


class FiscNetECF(SerialBase):
    implements(IChequePrinter, ICouponPrinter)

    log_domain = 'FiscNet'

    model_name = None
    coupon_printer_charset = "cp850"
    cheque_printer_charset = "ascii"
    supports_duplicate_receipt = True
    identify_customer_at_end = False

    CHEQUE_CONFIGFILE = 'perto.ini'

    CMD_PREFIX = '{'
    CMD_SUFFIX = '}'
    EOL_DELIMIT = CMD_SUFFIX

    CASH_SUPPLY = 'Suprimento'
    CASH_REMOVAL = 'Sangria'

    errors_dict = {
        7003: OutofPaperError,
        7004: OutofPaperError,
        8007: CouponTotalizeError,
        8011: PaymentAdditionError,
        8013: CouponTotalizeError,
        8014: PaymentAdditionError,
        8017: CloseCouponError,
        8044: CancelItemError,
        8045: CancelItemError,
        8068: PaymentAdditionError,
        8086: CancelItemError,
        15009: PendingReduceZ,
        11002: CommandParametersError,
        11006: CommandError,
        11007: InvalidState,
        15007: PendingReadX,
        15008: ReadXError,
        15011: OutofPaperError
    }

    def __init__(self, port, consts=None):
        port.setParity(PARITY_EVEN)
        port.setWriteTimeout(3)

        SerialBase.__init__(self, port)
        self._consts = consts or FiscNetConstants
        self._command_id = 0
        config = ConfigParser.RawConfigParser()
        has_conf = config.read([os.path.expanduser('~/.stoq/stoqdrivers.conf')])
        if has_conf:
            try:
                self.CASH_SUPPLY = config.get('fiscnet', 'cash_supply')
                self.CASH_REMOVAL = config.get('fiscnet', 'cash_removal')
            except ConfigParser.NoOptionError:
                log.info("Option not found, check stoqdrivers.conf")
        else:
            log.info("file: stoqdrivers.conf not found")
        self._reset()

    def _reset(self):
        self._customer_name = ''
        self._customer_document = ''
        self._customer_address = ''

    #
    # Helper methods
    #
    def _parse_return_value(self, text):

        # Based on cookielib.split_header_words
        def unmatched(match):
            start, end = match.span(0)
            return match.string[:start] + match.string[end:]

        result = {}
        while text:
            m = _RETVAL_TOKEN_RE.search(text)
            if m:
                text = unmatched(m)
                name = m.group(1)
                m = _RETVAL_QUOTED_VALUE_RE.search(text)
                if m:  # quoted value
                    text = unmatched(m)
                    value = m.group(1)
                    value = _RETVAL_ESCAPE_RE.sub(r"\1", value)
                else:
                    m = _RETVAL_VALUE_RE.search(text)
                    if m:  # unquoted value
                        text = unmatched(m)
                        value = m.group(1)
                        value = value.rstrip()
                    else:
                        # no value, a lone token
                        value = None
                result[name] = value
            else:
                raise AssertionError

        return result

    def _send_command(self, command, **params):
        # Page 38-39
        parameters = []
        for param, value in params.items():
            if isinstance(value, Decimal):
                value = ('%.03f' % value).replace('.', ',')
            elif isinstance(value, basestring):
                value = '"%s"' % value
            elif isinstance(value, bool):
                if value is False:
                    value = 'f'
                elif value is True:
                    value = 't'
            elif isinstance(value, datetime.date):
                value = value.strftime('#%d/%m/%y#')

            parameters.append('%s=%s' % (param, value))

        reply = self.writeline("%d;%s;%s;" % (self._command_id,
                                              command,
                                              ' '.join(parameters)))
        if reply[0] != '{':
            # This happened once after the first command issued after
            # the power returned, it should probably be handled gracefully
            raise AssertionError(repr(reply))

        # Page 39
        sections = reply[1:].split(';')
        if len(sections) != 4:
            raise AssertionError

        retdict = self._parse_return_value(sections[2])
        errorcode = int(sections[1])
        if errorcode != 0:
            errordesc = retdict['Circunstancia']
            try:
                exception = self.errors_dict[errorcode]
            except KeyError:
                raise DriverError(errordesc, errorcode)
            raise exception(errordesc, errorcode)

        return retdict

    def _read_register(self, name, regtype):
        if regtype == int:
            cmd = 'LeInteiro'
            argname = 'NomeInteiro'
            retname = 'ValorInteiro'
        elif regtype == Decimal:
            cmd = 'LeMoeda'
            argname = 'NomeDadoMonetario'
            retname = 'ValorMoeda'
        elif regtype == datetime.date:
            cmd = 'LeData'
            argname = 'NomeData'
            retname = 'ValorData'
        elif regtype == str:
            cmd = 'LeTexto'
            argname = 'NomeTexto'
            retname = 'ValorTexto'
        elif regtype == bool:
            cmd = 'LeIndicador'
            argname = 'NomeIndicador'
            retname = 'ValorNumericoIndicador'
        else:
            raise AssertionError

        retdict = self._send_command(cmd, **dict([(argname, name)]))
        assert retname in retdict
        retval = retdict[retname]
        if regtype == int:
            return int(retval)
        elif regtype == Decimal:
            retval = retval.replace('.', '')
            retval = retval.replace(',', '.')
            return Decimal(retval)
        elif regtype == datetime.date:
            # This happens the first time we send a ReducaoZ after
            # opening the printer and removing the jumper.
            if retval == '#00/00/0000#':
                return datetime.date.today()
            else:
                # "29/03/2007" -> datetime.date(2007, 3, 29)
                d, m, y = map(int, retval[1:-1].split('/'))
                return datetime.date(y, m, d)
        elif regtype == str:
            # '"string"' -> 'string'
            return retval.strip('"')
        elif regtype == bool:
            return bool(int(retval))
        else:
            raise AssertionError

    def _get_status(self):
        return self._read_register('Indicadores', int)

    def _get_last_item_id(self):
        return self._read_register('ContadorDocUltimoItemVendido', int)

    def _get_coupon_total_value(self):
        return self._read_register('TotalDocLiquido', Decimal)

    def _get_coupon_remainder_value(self):
        value = self._read_register('TotalDocValorPago', Decimal)
        result = self._get_coupon_total_value() - value
        if result < 0.0:
            result = 0.0
        return result

    # This how the printer needs to be configured.
    def _define_tax_name(self, code, name, entrada=False):
        try:
            retdict = self._send_command(
                'LeNaoFiscal', CodNaoFiscal=code)
        except DriverError, e:
            if e.code != 8057:  # Not configured
                raise
        else:
            for retname in ['NomeNaoFiscal', 'DescricaoNaoFiscal']:
                configured_name = retdict[retname]
                if configured_name != name:
                    raise DriverError(
                        "The name of the tax code %d is set to %r, "
                        "but it needs to be configured as %r" % (
                            code, configured_name, name))

        try:
            self._send_command(
                'DefineNaoFiscal', CodNaoFiscal=code, DescricaoNaoFiscal=name,
                NomeNaoFiscal=name, TipoNaoFiscal=entrada)
        except DriverError, e:
            if e.code != 8036:
                raise

    def _delete_tax_name(self, code):
        try:
            self._send_command(
                'ExcluiNaoFiscal', CodNaoFiscal=code)
        except DriverError, e:
            if e.code != 8057:  # Not configured
                raise

    def _define_payment_method(self, code, name, vinculated=False):
        try:
            retdict = self._send_command(
                'LeMeioPagamento', CodMeioPagamentoProgram=code)
        except DriverError, e:
            if e.code != 8014:  # Not configured
                raise
        else:
            configure = False
            for retname in ['NomeMeioPagamento', 'DescricaoMeioPagamento']:
                configured_name = retdict[retname]
                if configured_name != name:
                    configure = True

            if not configure:
                return

        try:
            self._send_command(
                'DefineMeioPagamento',
                CodMeioPagamentoProgram=code, DescricaoMeioPagamento=name,
                NomeMeioPagamento=name, PermiteVinculado=vinculated)
        except DriverError, e:
            raise

    def _delete_payment_method(self, code):
        try:
            self._send_command(
                'ExcluiMeioPagamento', CodMeioPagamentoProgram=code)
        except DriverError, e:
            if e.code != 8014:  # Not configured
                raise

    def _define_tax_code(self, code, value, service=False):
        try:
            retdict = self._send_command(
                'LeAliquota', CodAliquotaProgramavel=code)
        except DriverError, e:
            if e.code != 8005:  # Not configured
                raise
        else:
            configure = False
            for retname in ['PercentualAliquota']:
                configured_name = retdict[retname]
                if configured_name != value:
                    configure = True

            if not configure:
                return

        try:
            self._send_command(
                'DefineAliquota',
                CodAliquotaProgramavel=code,
                DescricaoAliquota='%2.2f%%' % value,
                PercentualAliquota=value,
                AliquotaICMS=not service)
        except DriverError, e:
            raise

    def _delete_tax_code(self, code):
        try:
            self._send_command(
                'ExcluiAliquota', CodAliquotaProgramavel=code)
        except DriverError, e:
            if e.code != 8005:  # Not configured
                raise

    def _get_taxes(self):
        taxes = []
        taxes.append(('CANC', Decimal(0), 'ICMS'))
        taxes.append(('DESC', Decimal(0), 'ICMS'))
        taxes.append(('I', Decimal(0), 'ICMS'))
        taxes.append(('N', Decimal(0), 'ICMS'))
        taxes.append(('F', Decimal(0), 'ICMS'))

        read_r = lambda x: self._read_register(x, Decimal)

        taxes = [
            ('I', read_r('TotalDiaIsencaoICMS'), 'ICMS'),
            ('F', read_r('TotalDiaSubstituicaoTributariaICMS'), 'ICMS'),
            ('N', read_r('TotalDiaNaoTributadoICMS'), 'ICMS'),
            ('DESC', read_r('TotalDiaDescontos'), 'ICMS'),
            ('CANC', read_r('TotalDiaCancelamentosICMS'), 'ICMS'),
        ]

        for reg in range(16):
            value = read_r('TotalDiaValorAliquota[%d]' % reg)
            if value:
                retdict = self._send_command(
                    'LeAliquota', CodAliquotaProgramavel=reg)

                if retdict['AliquotaICMS'] == 'Y':
                    type = "ICMS"
                else:
                    type = "ISS"

                desc = retdict['PercentualAliquota'].replace(',', '')
                taxes.append(('%04d' % int(desc), value, type))
        return taxes

    def _configure_printer(self):
        self._define_tax_name(0,
                              self.CASH_SUPPLY.encode(self.coupon_printer_charset), entrada=True)
        self._define_tax_name(1,
                              self.CASH_REMOVAL.encode(self.coupon_printer_charset), entrada=False)
        for code in range(2, 15):
            self._delete_tax_name(code)

        self._define_payment_method(0,
                                    u'Cheque'.encode(self.coupon_printer_charset))
        self._define_payment_method(1,
                                    u'Boleto'.encode(self.coupon_printer_charset))
        self._define_payment_method(2,
                                    u'Cartão credito'.encode(self.coupon_printer_charset),
                                    vinculated=True)
        self._define_payment_method(3,
                                    u'Cartão debito'.encode(self.coupon_printer_charset),
                                    vinculated=True)
        self._define_payment_method(4,
                                    u'Financeira'.encode(self.coupon_printer_charset))
        self._define_payment_method(5,
                                    u'Vale compra'.encode(self.coupon_printer_charset))
        for code in range(6, 15):
            self._delete_payment_method(code)

        self._define_tax_code(0, Decimal("17.00"))
        self._define_tax_code(1, Decimal("12.00"))
        self._define_tax_code(2, Decimal("25.00"))
        self._define_tax_code(3, Decimal("8.00"))
        self._define_tax_code(4, Decimal("5.00"))
        self._define_tax_code(5, Decimal("3.00"), service=True)
        for code in range(6, 16):
            self._delete_tax_code(code)

    def print_status(self):
        status = self._get_status()
        print 'Flags'
        for flag in reversed(_status_flags):
            if status & flag:
                print flag, _flagnames[flag]

        print 'non-fiscal registers'
        for i in range(15):
            try:
                print self._send_command(
                    'LeNaoFiscal', CodNaoFiscal=i)
            except DriverError, e:
                if e.code != 8057:
                    raise

    #
    # ICouponPrinter implementation
    #

    def setup(self):
        pass

    def coupon_identify_customer(self, customer, address, document):
        self._customer_name = customer
        self._customer_document = document
        self._customer_address = address

    def coupon_is_customer_identified(self):
        return len(self._customer_document) > 0

    def has_open_coupon(self):
        return self._read_register('DocumentoAberto', bool)

    def coupon_open(self):
        status = self._get_status()
        if status & FLAG_DOCUMENTO_ABERTO:
            raise CouponOpenError(_("Coupon already opened."))

        customer = self._customer_name
        document = self._customer_document
        address = self._customer_address
        self._send_command('AbreCupomFiscal',
                           EnderecoConsumidor=address[:80],
                           IdConsumidor=document[:29],
                           NomeConsumidor=customer[:30])

    def coupon_add_item(self, code, description, price, taxcode,
                        quantity=Decimal("1.0"), unit=UnitType.EMPTY,
                        discount=Decimal("0.0"), surcharge=Decimal("0.0"),
                        unit_desc=""):
        status = self._get_status()
        if not status & FLAG_DOCUMENTO_ABERTO:
            raise CouponNotOpenError

        if unit == UnitType.CUSTOM:
            unit = unit_desc
        else:
            unit = self._consts.get_value(unit)

        taxcode = int(taxcode)
        self._send_command('VendeItem',
                           CodAliquota=taxcode,
                           CodProduto=code[:48],
                           NomeProduto=description[:200],
                           Unidade=unit,
                           PrecoUnitario=price,
                           Quantidade=quantity)

        if discount:
            self._send_command('AcresceItemFiscal',
                               Cancelar=False,
                               ValorAcrescimo=-discount)

        return self._get_last_item_id()

    def coupon_cancel_item(self, item_id):
        self._send_command('CancelaItemFiscal', NumItem=item_id)

    def coupon_cancel(self):
        self._send_command('CancelaCupom')

    def cancel_last_coupon(self):
        """Cancel the last non fiscal coupon or the last sale."""
        self._send_command('CancelaCupom')

    def coupon_totalize(self, discount=currency(0),
                        surcharge=currency(0), taxcode=TaxType.NONE):
        # The FISCnet protocol (the protocol used in this printer model)
        # doesn't have a command to totalize the coupon, so we just get
        # the discount/surcharge values and applied to the coupon.

        value = discount and (discount * -1) or surcharge
        if value:
            self._send_command('AcresceSubtotal',
                               Cancelar=False,
                               ValorAcrescimo=value)
        return self._get_coupon_total_value()

    def coupon_add_payment(self, payment_method, value, description=u""):
        pm = int(payment_method)
        self._send_command('PagaCupom',
                           CodMeioPagamento=pm, Valor=value,
                           TextoAdicional=description[:80])
        return self._get_coupon_remainder_value()

    def coupon_close(self, message=''):
        self._send_command('EncerraDocumento',
                           TextoPromocional=message[:492])
        self._reset()
        return self.get_coo()

    def summarize(self):
        self._send_command('EmiteLeituraX')

    def open_till(self):
        self.summarize()

    def close_till(self, previous_day=False):
        status = self._get_status()
        if status & FLAG_DOCUMENTO_ABERTO:
            self.coupon_cancel()

        self._send_command('EmiteReducaoZ')

    def has_pending_reduce(self):
        status = self._get_status()
        return status & FLAG_Z_PENDENTE

    def till_add_cash(self, value):
        status = self._get_status()
        if status & FLAG_DOCUMENTO_ABERTO:
            self.coupon_cancel()
        self._send_command('AbreCupomNaoFiscal')
        self._send_command('EmiteItemNaoFiscal',
                           NomeNaoFiscal=self.CASH_SUPPLY,
                           Valor=value)
        self._send_command('PagaCupom',
                           CodMeioPagamento=-2, Valor=value)
        self._send_command('EncerraDocumento')

    def till_remove_cash(self, value):
        status = self._get_status()
        if status & FLAG_DOCUMENTO_ABERTO:
            self.coupon_cancel()
        self._send_command('AbreCupomNaoFiscal')
        self._send_command('EmiteItemNaoFiscal',
                           NomeNaoFiscal=self.CASH_REMOVAL,
                           Valor=value)
        self._send_command('EncerraDocumento')

    def till_read_memory(self, start=None, end=None):
        try:
            self._send_command('EmiteLeituraMF',
                               LeituraSimplificada=True,
                               DataInicial=start,
                               DataFinal=end)
        except DriverError, e:
            if e.code == 8089:
                return

    def till_read_memory_by_reductions(self, start=None, end=None):
        self._send_command('EmiteLeituraMF',
                           LeituraSimplificada=True,
                           ReducaoInicial=start,
                           ReducaoFinal=end)

    def get_capabilities(self):
        return dict(item_code=Capability(max_len=48),
                    item_id=Capability(max_size=32767),
                    items_quantity=Capability(digits=14, decimals=4),
                    item_price=Capability(digits=14, decimals=4),
                    item_description=Capability(max_len=200),
                    payment_value=Capability(digits=14, decimals=4),
                    promotional_message=Capability(max_len=492),
                    payment_description=Capability(max_len=80),
                    customer_name=Capability(max_len=30),
                    customer_id=Capability(max_len=29),
                    customer_address=Capability(max_len=80))

    def get_constants(self):
        return self._consts

    def query_status(self):
        return '{0;LeInteiro;NomeInteiro="Indicadores";}'

    def status_reply_complete(self, reply):
        return '}' in reply

    def get_serial(self):
        return self._read_register('NumeroSerieECF', str)

    def get_tax_constants(self):
        constants = []

        for reg in range(16):
            try:
                retdict = self._send_command(
                    'LeAliquota', CodAliquotaProgramavel=reg)
            except DriverError, e:
                if e.code == 8005:  # Aliquota nao carregada
                    continue
                raise

            if retdict['AliquotaICMS'] == 'Y':
                tax_type = TaxType.CUSTOM
            else:
                tax_type = TaxType.SERVICE

            value = Decimal(retdict['PercentualAliquota'].replace(',', '.'))
            device_value = retdict['CodAliquotaProgramavel']
            constants.append((tax_type, device_value, value))

        # FIXME
        # These are signed integers, we're storing them
        # as strings and then subtract by 127
        # Page 10
        constants.extend([
            (TaxType.SUBSTITUTION, '-2', None),  # -2
            (TaxType.EXEMPTION, '-3', None),  # -3
            (TaxType.NONE, '-4', None),  # -4
        ])

        return constants

    def get_payment_constants(self):
        constants = [('-2', u'Dinheiro')]

        for reg in range(16):
            try:
                retdict = self._send_command(
                    'LeMeioPagamento', CodMeioPagamentoProgram=reg)
            except DriverError, e:
                if e.code == 8014:  # Meio de pagamento nao carregado
                    continue
                raise

            code = retdict['CodMeioPagamentoProgram']
            name = retdict['NomeMeioPagamento']
            constants.append((code, name.decode(self.coupon_printer_charset)))

        return constants

    def get_ccf(self):
        return self._read_register('CCF', int)

    def get_coo(self):
        return self._read_register('COO', int)

    def get_gnf(self):
        return self._read_register('GNF', int)

    def get_cro(self):
        return self._read_register('CRO', int)

    def get_crz(self):
        return self._read_register('CRZ', int)

    def get_firmware_version(self):
        return self._read_register('VersaoSW', str)

    def get_sintegra(self):
        data = Settable(
            opening_date=self._read_register('DataAbertura', datetime.date),
            serial=self.get_serial(),
            serial_id=self._read_register('ECF', int),
            coupon_start=self._read_register('COOInicioDia', int),
            coupon_end=self.get_coo(),
            cro=self.get_cro(),
            crz=self.get_crz(),
            coo=self.get_coo(),
            period_total=self._read_register('TotalDiaVendaBruta', Decimal),
            total=self._read_register('GT', Decimal),
            taxes=self._get_taxes())

        return data

    def get_payment_receipt_identifier(self, method_name):
        # We don't need an identifier. Just return None
        return None

    def payment_receipt_open(self, identifier, coo, method_id, value):
        self._send_command('AbreCreditoDebito', CodMeioPagamento=int(method_id),
                           COO=coo, Valor=value)

    def payment_receipt_print(self, text):
        text = text.encode(self.coupon_printer_charset)
        for line in text.split('\n'):
            line = line.replace('\\', '\\\\')  # Vespague sucks
            self._send_command('ImprimeTexto', TextoLivre=line)

    def payment_receipt_close(self):
        self._send_command('EncerraDocumento')

    def payment_receipt_print_duplicate(self):
        self._send_command('EmiteViaCreditoDebito')

    def gerencial_report_open(self, gerencial_id=0):
        self._send_command('AbreGerencial', CodGerencial=gerencial_id)

    def gerencial_report_print(self, text):
        text = text.encode(self.coupon_printer_charset)
        for line in text.split('\n'):
            line = line.replace('\\', '\\\\')  # Vespague sucks
            self._send_command('ImprimeTexto', TextoLivre=line)

    def gerencial_report_close(self):
        self._send_command('EncerraDocumento')


class FiscNetChequePrinter(FiscNetECF, BaseChequePrinter):
    implements(IChequePrinter)

    def __init__(self, port, consts=None):
        FiscNetECF.__init__(self, port, consts)
        BaseChequePrinter.__init__(self)

    #
    # IChequePrinter implementation
    #

    def print_cheque(self, bank, value, thirdparty, city, date=None):
        if date is None:
            data = datetime.datetime.now()
        if not isinstance(bank, BankConfiguration):
            raise TypeError("bank parameter must be BankConfiguration instance")

        data = dict(HPosAno=bank.get_x_coordinate("year"),
                    HPosCidade=bank.get_x_coordinate("city"),
                    HPosDia=bank.get_x_coordinate("day"),
                    HPosExtensoLinha1=bank.get_x_coordinate("legal_amount"),
                    HPosExtensoLinha2=bank.get_x_coordinate("legal_amount2"),
                    HPosFavorecido=bank.get_x_coordinate("thirdparty"),
                    HPosMes=bank.get_x_coordinate("month"),
                    HPosValor=bank.get_x_coordinate("value"),
                    VPosCidade=bank.get_y_coordinate("city"),
                    VPosExtensoLinha1=bank.get_y_coordinate("legal_amount"),
                    VPosExtensoLinha2=bank.get_y_coordinate("legal_amount2"),
                    VPosFavorecido=bank.get_y_coordinate("thirdparty"),
                    VPosValor=bank.get_y_coordinate("value"))

        self._send_command('ImprimeCheque', Cidade=city[:27],
                           Data=date.strftime("#%d/%m/%Y#"),
                           Favorecido=thirdparty[:45],
                           Valor=value, **data)

    def _get_cheque_capabilities(self):
        return dict(cheque_thirdparty=Capability(max_len=45),
                    cheque_value=Capability(digits=14, decimals=4),
                    cheque_city=Capability(max_len=27))

    def get_capabilities(self):
        caps = super(FiscNetChequePrinter, self).get_capabilities()
        caps.update(self._get_cheque_capabilities())
        return caps



# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2005 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Henrique Romano <henrique@async.com.br>
##
"""
Driver Capability management.
"""

from kiwi.argcheck import argcheck, number

from stoqdrivers.exceptions import CapabilityError


class capcheck(argcheck):
    """ A extension for argcheck that validates a value with base in the driver
    capabilities.  Note that the instance where this class is used as decorator
    must have defined a get_capabilities  method that returns a dictionary with
    the driver capabilities.
    """

    def extra_check(self, arg_names, types, cargs, kwargs):
        keyvalues = zip(arg_names, cargs[1:])
        kwargs = kwargs.copy()
        kwargs.update(dict(keyvalues))
        self._check_capabilities(cargs[0], **kwargs)

    def _check_capabilities(self, inst, **kwargs):
        caps = inst.get_capabilities()

        for key, value in kwargs.items():
            capability = caps.get(key)
            if not capability:
                continue
            try:
                capability.check_value(value)
            except CapabilityError, e:
                raise CapabilityError("invalid value for '%s': %s" % (key, e))


class Capability:
    """ This class is used to represent a driver capability, offering methods
    to validate a value with base in the capability limits.
    """

    @argcheck(int, int, number, number, int, number)
    def __init__(self, min_len=None, max_len=None, max_size=None,
                 min_size=None, digits=None, decimals=None):
        """ Creates a new driver capability.  A driver capability can be
        represented basically by the max length of a string, the max digits
        number of a value or its minimum/maximum size.  With an instance of
        Capability you can check if a value is acceptable by the driver
        through the check_value method.  The Capability arguments are:

        @param min_len:    The minimum length of a string
        @type min_len:     number
        @param max_len:    The max length of a string
        @type max_len:     number
        @param max_size    The maximum size for a value
        @type max_size:    number
        @param min_size:   The minimum size for a value
        @type min_size:    number
        @param digits:     The number of digits that a number can have
        @type digits:      number
        @param decimals:   If the max value for the capability is a float,
                           this parameter specifies the max precision
                           that the number can have.
        @type decimals:    number

        Note that 'max_len' can't be used together with 'min_size', 'max_size'
        and 'digits', in the same way that 'max_size' and 'min_size' can't be
        used with 'digits'.  The values defined for these parameters are used
        also to verify the value type in the 'check_value' method.
        """

        if max_len is not None and (max_size is not None
                                    and min_size is not None
                                    and digits is not None
                                    and decimals):
            raise ValueError("max_len cannot be used together with max_size, "
                             "min_size, digits or decimals")
        if digits is not None:
            if max_size is not None:
                raise ValueError("digits can't be used with max_size")
            if decimals:
                decimal_part = 1 - (1 / 10.0 ** decimals)
            else:
                decimal_part = 0
            self.max_size = ((10.0 ** digits) - 1) + decimal_part

        self.min_len = min_len
        self.max_len = max_len
        self.min_size = min_size or 0
        self.max_size = max_size
        self.digits = digits
        self.decimals = decimals

    def check_value(self, value):
        if self.max_len:
            if not isinstance(value, basestring):
                raise CapabilityError("the value must be a string")
            if len(value) > self.max_len:
                raise CapabilityError("the value can't be greater than %d "
                                      "characters" % self.max_len)
            elif len(value) < self.min_len:
                raise CapabilityError("the value can't be less than %d "
                                      "characters" % self.min_len)
            return
        elif not (self.max_size and self.min_size):
            return

        if not isinstance(value, (float, int, long)):
            raise CapabilityError("the value must be float, integer or long")

        if value > self.max_size:
            raise CapabilityError("the value can't be greater than %r"
                                  % self.max_size)
        elif value < self.min_size:
            raise CapabilityError("the value can't be less than %r"
                                  % self.min_size)



# -*- Mode: Python; coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2006 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Henrique Romano  <henrique@async.com.br>
##

from kiwi.python import namedAny
from zope.interface import implements

from stoqdrivers.interfaces import IBarcodeReader
from stoqdrivers.serialbase import SerialBase


class BaseBarcodeReader(SerialBase):
    implements(IBarcodeReader)
    # Should be defined in subclasses
    model_name = None

    def __init__(self, port, consts=None):
        SerialBase.__init__(self, port)

    def get_code(self):
        return self.readline()


def get_supported_barcode_readers():
    result = {}
    for brand, module_names in [('metrologic', ['MC630'])]:
        result[brand] = []
        for module_name in module_names:
            try:
                obj = namedAny("stoqdrivers.readers.barcode.%s.%s.%s"
                               % (brand, module_name, module_name))
            except AttributeError:
                raise ImportError("Can't find class %s for module %s"
                                  % (module_name, module_name))
            if not IBarcodeReader.implementedBy(obj):
                raise TypeError("The driver %s %s doesn't implements a "
                                "valid interface"
                                % (brand, obj.model_name))
            result[brand].append(obj)
    return result



# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2012 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s): Stoq Team <stoq-devel@async.com.br>
##
"""
Epson FBIII ECF driver
"""
from decimal import Decimal

from stoqdrivers.printers.epson.FBII import FBII

# Coupons status
CLOSED_COUPON = '0000'
OPENED_FISCAL_COUPON = '0001'
OPENED_NON_FISCAL_COUPON = '1000'


class FBIII(FBII):
    model_name = "Epson FBIII"

    def apply_discount(self, id, discount):
        value = int(discount * Decimal('1e2'))
        self._send_command('0A07', '0010', id, str(value))

    def apply_markup(self, id, markup):
        value = int(markup * Decimal('1e2'))
        self._send_command('0A07', '0011', id, str(value))



# -*- Mode: Python; coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2005-2007 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Johan Dahlin     <jdahlin@async.com.br>
##              Henrique Romano  <henrique@async.com.br>
##

import datetime
from decimal import Decimal
import logging

from kiwi.argcheck import number
from kiwi.currency import currency

from stoqdrivers.exceptions import (CloseCouponError, PaymentAdditionError,
                                    AlreadyTotalized, InvalidValue)
from stoqdrivers.enum import PaymentMethodType, TaxType, UnitType
from stoqdrivers.printers.base import BasePrinter
from stoqdrivers.printers.capabilities import capcheck
from stoqdrivers.utils import encode_text
from stoqdrivers.translation import stoqdrivers_gettext

_ = stoqdrivers_gettext

log = logging.getLogger('stoqdrivers.fiscalprinter')

#
# Extra data types to argcheck
#


class taxcode(number):
    @classmethod
    def value_check(cls, name, value):
        if value not in (TaxType.NONE, TaxType.ICMS, TaxType.SUBSTITUTION,
                         TaxType.EXEMPTION):
            raise ValueError("%s must be one of TaxType.* constants" % name)


class unit(number):
    @classmethod
    def value_check(cls, name, value):
        if value not in (UnitType.WEIGHT, UnitType.METERS, UnitType.LITERS,
                         UnitType.EMPTY, UnitType.CUSTOM):
            raise ValueError("%s must be one of UNIT_* constants" % name)


class payment_method(number):
    @classmethod
    def value_check(cls, name, value):
        if value not in (PaymentMethodType.MONEY, PaymentMethodType.CHECK,
                         PaymentMethodType.CUSTOM):
            raise ValueError("%s must be one of *_PM constants" % name)

#
# FiscalPrinter interface
#


class FiscalPrinter(BasePrinter):
    def __init__(self, brand=None, model=None, device=None, config_file=None,
                 *args, **kwargs):
        BasePrinter.__init__(self, brand, model, device, config_file, *args,
                             **kwargs)
        self._has_been_totalized = False
        self.payments_total_value = Decimal("0.0")
        self.totalized_value = Decimal("0.0")
        self._capabilities = self._driver.get_capabilities()
        self._charset = self._driver.coupon_printer_charset
        try:
            self.setup()
            self._setup_complete = True
        except Exception:
            self._setup_complete = False

    def setup_complete(self):
        return self._setup_complete

    def get_capabilities(self):
        return self._capabilities

    def _format_text(self, text):
        return encode_text(text, self._charset)

    def setup(self):
        log.info('setup()')
        self._driver.setup()

    @capcheck(basestring, basestring, basestring)
    def identify_customer(self, customer_name, customer_address, customer_id):
        log.info('identify_customer(customer_name=%r, '
                 'customer_address=%r, customer_id=%r)' % (
                     customer_name, customer_address, customer_id))

        self._driver.coupon_identify_customer(
            self._format_text(customer_name),
            self._format_text(customer_address),
            self._format_text(customer_id))

    def coupon_is_customer_identified(self):
        return self._driver.coupon_is_customer_identified()

    def has_open_coupon(self):
        log.info('has_open_coupon()')
        return self._driver.has_open_coupon()

    def open(self):
        log.info('coupon_open()')

        return self._driver.coupon_open()

    @capcheck(basestring, basestring, Decimal, str, Decimal, unit,
              Decimal, Decimal, basestring)
    def add_item(self, item_code, item_description, item_price, taxcode,
                 items_quantity=Decimal("1.0"), unit=UnitType.EMPTY,
                 discount=Decimal("0.0"), surcharge=Decimal("0.0"),
                 unit_desc=""):
        log.info("add_item(code=%r, description=%r, price=%r, "
                 "taxcode=%r, quantity=%r, unit=%r, discount=%r, "
                 "surcharge=%r, unit_desc=%r)" % (
                     item_code, item_description, item_price, taxcode,
                     items_quantity, unit, discount, surcharge, unit_desc))

        if self._has_been_totalized:
            raise AlreadyTotalized("the coupon is already totalized, you "
                                   "can't add more items")
        if discount and surcharge:
            raise TypeError("discount and surcharge can not be used together")
        elif unit != UnitType.CUSTOM and unit_desc:
            raise ValueError("You can't specify the unit description if "
                             "you aren't using UnitType.CUSTOM constant.")
        elif unit == UnitType.CUSTOM and not unit_desc:
            raise ValueError("You must specify the unit description when "
                             "using UnitType.CUSTOM constant.")
        elif unit == UnitType.CUSTOM and len(unit_desc) != 2:
            raise ValueError("unit description must be 2-byte sized string")
        if not item_price:
            raise InvalidValue("The item value must be greater than zero")

        if surcharge < 0:
            raise ValueError('Surcharge cannot be negative')
        if discount < 0:
            raise ValueError('Discount cannot be negative')

        return self._driver.coupon_add_item(
            self._format_text(item_code), self._format_text(item_description),
            item_price, taxcode, items_quantity, unit, discount, surcharge,
            unit_desc=self._format_text(unit_desc))

    @capcheck(Decimal, Decimal, taxcode)
    def totalize(self, discount=currency(0), surcharge=currency(0),
                 taxcode=TaxType.NONE):
        log.info('totalize(discount=%r, surcharge=%r, taxcode=%r)' % (
            discount, surcharge, taxcode))

        if discount and surcharge:
            raise TypeError("discount and surcharge can not be used together")
        if surcharge and taxcode == TaxType.NONE:
            raise ValueError("to specify a surcharge you need specify its "
                             "tax code")
        result = self._driver.coupon_totalize(discount, surcharge, taxcode)
        self._has_been_totalized = True
        self.totalized_value = result
        return result

    @capcheck(basestring, Decimal, basestring)
    def add_payment(self, payment_method, payment_value, description=''):
        log.info("add_payment(method=%r, value=%r, description=%r)" % (
            payment_method, payment_value, description))

        if not self._has_been_totalized:
            raise PaymentAdditionError(_("You must totalize the coupon "
                                         "before add payments."))
        result = self._driver.coupon_add_payment(
            payment_method, payment_value,
            self._format_text(description))
        self.payments_total_value += payment_value
        return result

    def cancel(self):
        log.info('coupon_cancel()')
        retval = self._driver.coupon_cancel()
        self._has_been_totalized = False
        self.payments_total_value = Decimal("0.0")
        self.totalized_value = Decimal("0.0")
        return retval

    def cancel_last_coupon(self):
        """Cancel the last non fiscal coupon or the last sale."""
        log.info('cancel_last_coupon()')
        self._driver.cancel_last_coupon()

    @capcheck(int)
    def cancel_item(self, item_id):
        log.info('coupon_cancel_item(item_id=%r)' % (item_id,))

        return self._driver.coupon_cancel_item(item_id)

    @capcheck(basestring)
    def close(self, promotional_message=''):
        log.info('coupon_close(promotional_message=%r)' % (
            promotional_message))

        if not self._has_been_totalized:
            raise CloseCouponError(_("You must totalize the coupon before "
                                     "closing it"))
        if not self.payments_total_value:
            raise CloseCouponError(_("It is not possible close the coupon "
                                     "since there are no payments defined."))
        if self.totalized_value > self.payments_total_value:
            raise CloseCouponError(_("Isn't possible close the coupon since "
                                     "the payments total (%.2f) doesn't "
                                     "match the totalized value (%.2f).")
                                   % (self.payments_total_value,
                                      self.totalized_value))
        res = self._driver.coupon_close(
            self._format_text(promotional_message))
        self._has_been_totalized = False
        self.payments_total_value = Decimal("0.0")
        self.totalized_value = Decimal("0.0")
        return res

    def summarize(self):
        log.info('summarize()')

        return self._driver.summarize()

    def has_pending_reduce(self):
        pending = self._driver.has_pending_reduce()
        log.info('has_pending_reduce() = %s' % pending)
        return pending

    def open_till(self):
        log.info('open_till()')
        return self._driver.open_till()

    def close_till(self, previous_day=False):
        log.info('close_till(previous_day=%r)' % (previous_day,))

        return self._driver.close_till(previous_day)

    @capcheck(Decimal)
    def till_add_cash(self, add_cash_value):
        log.info('till_add_cash(add_cash_value=%r)' % (add_cash_value,))

        return self._driver.till_add_cash(add_cash_value)

    @capcheck(Decimal)
    def till_remove_cash(self, remove_cash_value):
        log.info('till_remove_cash(remove_cash_value=%r)' % (
            remove_cash_value,))

        return self._driver.till_remove_cash(remove_cash_value)

    @capcheck(datetime.date, datetime.date)
    def till_read_memory(self, start, end):
        assert start <= end <= datetime.date.today(), (
            "start must be less then end and both must be less today")
        log.info('till_read_memory(start=%r, end=%r)' % (
            start, end))

        return self._driver.till_read_memory(start, end)

    @capcheck(datetime.date, datetime.date)
    def till_read_memory_to_serial(self, start, end):
        assert start <= end <= datetime.date.today(), (
            "start must be less then end and both must be less today")
        log.info('till_read_memory(start=%r, end=%r)' % (
            start, end))

        return self._driver.till_read_memory_to_serial(start, end)

    @capcheck(int, int)
    def till_read_memory_by_reductions(self, start, end):
        assert end >= start > 0, ("start must be less then end "
                                  "and both must be positive")
        log.info('till_read_memory_by_reductions(start=%r, end=%r)' % (
            start, end))

        self._driver.till_read_memory_by_reductions(start, end)

    def gerencial_report_open(self):
        log.info('gerencial_report_open')
        return self._driver.gerencial_report_open()

    def gerencial_report_print(self, text):
        log.info('gerencial_report_print(text=%s)' % text)
        return self._driver.gerencial_report_print(text)

    def gerencial_report_close(self):
        log.info('gerencial_report_close')
        return self._driver.gerencial_report_close()

    def payment_receipt_open(self, identifier, coo, method, value):
        log.info('payment_receipt_open(identifier=%s, coo=%s, method=%s, value=%s)'
                 % (identifier, coo, method, value))
        return self._driver.payment_receipt_open(identifier, coo, method, value)

    def payment_receipt_print(self, text):
        log.info('payment_receipt_print(text=%s)' % text)
        return self._driver.payment_receipt_print(text)

    def payment_receipt_close(self):
        log.info('payment_receipt_close()')
        return self._driver.payment_receipt_close()

    def payment_receipt_print_duplicate(self):
        log.info('payment_receipt_print_duplicate()')
        return self._driver.payment_receipt_print_duplicate()

    def get_serial(self):
        log.info('get_serial()')

        return self._driver.get_serial()

    def query_status(self):
        log.info('query_status()')

        return self._driver.query_status()

    def status_reply_complete(self, reply):
        log.info('status_reply_complete(%s)' % (reply,))
        return self._driver.status_reply_complete(reply)

    def get_tax_constants(self):
        log.info('get_tax_constants()')

        return self._driver.get_tax_constants()

    def get_payment_constants(self):
        log.info('get_payment_constants()')

        return self._driver.get_payment_constants()

    def get_payment_receipt_identifier(self, method):
        log.info('get_payment_receipt_identifier(method=%s)' % method)
        return self._driver.get_payment_receipt_identifier(method)

    def get_ccf(self):
        """Fiscal Coupon Counter

        @returns: the last document ccf
        @rtype: integer
        """
        log.info('get_ccf()')

        return self._driver.get_ccf()

    def get_coo(self):
        """Operation Order Counter

        @returns: the last document coo
        @rtype: integer
        """
        log.info('get_coo()')

        return self._driver.get_coo()

    def get_gnf(self):
        """Nonfiscal Operation General Counter

        @returns: gnf
        @rtype: integer
        """
        log.info('get_gnf()')

        return self._driver.get_gnf()

    def get_crz(self):
        """Z Reduction Counter

        @returns: the last document crz
        @rtype: integer
        """
        log.info('get_crz()')

        return self._driver.get_crz()

    def get_sintegra(self):
        log.info('get_sintegra()')

        return self._driver.get_sintegra()

    @property
    def supports_duplicate_receipt(self):
        return self._driver.supports_duplicate_receipt

    @property
    def identify_customer_at_end(self):
        return self._driver.identify_customer_at_end



# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2006 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s): Henrique Romano             <henrique@async.com.br>
##
"""
Generic base class implementation for all devices.
"""

import logging

import gobject

from stoqdrivers.configparser import StoqdriversConfig
from stoqdrivers.enum import DeviceType
from stoqdrivers.exceptions import CriticalError, ConfigError
from stoqdrivers.translation import stoqdrivers_gettext
from stoqdrivers.serialbase import SerialPort

_ = stoqdrivers_gettext

log = logging.getLogger('stoqdrivers.basedev')


class BaseDevice:
    """ Base class for all device interfaces, responsible for instantiate
    the device driver itself based on the brand and model specified or in
    the configuration file.
    """
    typename_translate_dict = {
        DeviceType.PRINTER: "Printer",
        DeviceType.SCALE: "Scale",
        DeviceType.BARCODE_READER: "Barcode Reader",
    }
    # Subclasses must define these attributes
    device_dirname = None
    required_interfaces = None
    device_type = None

    def __init__(self, brand=None, model=None, device=None, config_file=None,
                 port=None, consts=None):
        if not self.device_dirname:
            raise ValueError("Subclasses must define the "
                             "`device_dirname' attribute")
        elif self.device_type is None:
            raise ValueError("device_type must be defined")
        self.brand = brand
        self.device = device
        self.model = model
        self._port = port
        self._driver_constants = consts
        self._load_configuration(config_file)

    def _load_configuration(self, config_file):
        section_name = BaseDevice.typename_translate_dict[self.device_type]
        if not self.model or not self.brand or (not self.device and not self._port):
            self.config = StoqdriversConfig(config_file)
            if not self.config.has_section(section_name):
                raise ConfigError(_("There is no section named `%s'!")
                                  % section_name)
            self.brand = self.config.get_option("brand", section_name)
            self.device = self.config.get_option("device", section_name)
            self.model = self.config.get_option("model", section_name)

        name = "stoqdrivers.%s.%s.%s" % (self.device_dirname,
                                         self.brand, self.model)
        try:
            module = __import__(name, None, None, 'stoqdevices')
        except ImportError, reason:
            raise CriticalError("Could not load driver %s %s: %s"
                                % (self.brand.capitalize(),
                                   self.model.upper(), reason))
        class_name = self.model
        driver_class = getattr(module, class_name, None)
        if not driver_class:
            raise CriticalError("Device driver at %s needs a class called %s"
                                % (name, class_name))
        if not self._port:
            self._port = SerialPort(self.device)

        self._driver = driver_class(self._port, consts=self._driver_constants)
        log.info(("Config data: brand=%s,device=%s,model=%s"
                  % (self.brand, self.device, self.model)))
        self.check_interfaces()

    def get_model_name(self):
        return self._driver.model_name

    def get_firmware_version(self):
        """Printer firmware version
        """
        return self._driver.get_firmware_version()

    def check_interfaces(self):
        """ This method must be implemented in subclass and must ensure that the
        driver implements a valid interface for the current operation state.
        """
        raise NotImplementedError

    def notify_read(self, func):
        """ This function can be called when the callsite must know when data
        is coming from the serial port.   It is necessary that a gobject main
        loop is already running before calling this method.
        """
        gobject.io_add_watch(self.get_port().fd, gobject.IO_IN,
                             lambda fd, cond: func(self, cond))

    def set_port(self, port):
        self._driver.set_port(port)

    def get_port(self):
        return self._driver.get_port()



# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2005 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Henrique Romano  <henrique@async.com.br>
##
"""
Implementation of Toled Prix III driver.
"""

from zope.interface import implements

from stoqdrivers.exceptions import InvalidReply
from stoqdrivers.interfaces import IScale, IScaleInfo
from stoqdrivers.serialbase import SerialBase, SerialPort

STX = 0x02
ETX = 0x03

PRICE_PRECISION = 2
QUANTITY_PRECISION = 3


class PackagePrt4:
    """ This class implements a parser for the 4a protocol of Toledo Prix III

    This protocol requires the user to press the print button.
    """
    implements(IScaleInfo)

    SIZE = 25

    def __init__(self, raw_data):
        self.code = None
        self.price_per_kg = None
        self.total_price = None
        self.weight = None
        self._parse(raw_data)

    def _parse(self, data):
        if not data:
            return
        elif ord(data[0]) != STX or len(data) != self.SIZE:
            raise InvalidReply("Received inconsistent data")
        self.code = int(data[1:7])
        self.price_per_kg = float(data[12:18]) / (10 ** PRICE_PRECISION)
        self.weight = float(data[7:12]) / (10 ** QUANTITY_PRECISION)
        self.total_price = float(data[18:24]) / (10 ** PRICE_PRECISION)


class PackagePrt1:
    """ This class implements a parser for the protocol prt1 of Toledo Prix III

    this protocol does not require any interaction from the user. The driver
    queries the scale for the weight, and the scale replies
    """
    implements(IScaleInfo)
    SIZE = 7

    def __init__(self, raw_data):
        self.code = None
        self.price_per_kg = None
        self.total_price = None
        self.weight = None
        self._parse(raw_data)

    def _parse(self, data):
        if not data:
            return
        elif ord(data[0]) != STX or len(data) != self.SIZE:
            raise InvalidReply("Received inconsistent data")
        self.weight = float(data[1:6]) / (10 ** QUANTITY_PRECISION)


class PrixIII(SerialBase):
    CMD_PREFIX = "\x05"
    EOL_DELIMIT = chr(ETX)

    implements(IScale)

    model_name = "Toledo Prix III"

    def __init__(self, device, consts=None):
        SerialBase.__init__(self, device)
        self._package = None

    def _get_package(self):
        reply = self.writeline('')
        # The sum is just because readline (called internally by writeline)
        # remove the EOL_DELIMIT from the package received and we need send
        # to Package's constructor the whole data.
        return PackagePrt1(reply + PrixIII.EOL_DELIMIT)

    #
    # IScale implementation
    #

    def read_data(self):
        return self._get_package()


if __name__ == "__main__":
    port = SerialPort('/dev/ttyS0')
    r = PrixIII(port)
    data = r.read_data()

    print "WEIGHT:", data.weight
    print "PRICE BY KG:", data.price_per_kg
    print "TOTAL PRICE:", data.total_price
    print "CODE:", data.code



# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2005 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s): Henrique Romano             <henrique@async.com.br>
##
"""
Useful functions related to all scales supported by stoqdrivers
"""

from kiwi.python import namedAny
from zope.interface import providedBy

from stoqdrivers.interfaces import IScale
from stoqdrivers.base import BaseDevice
from stoqdrivers.enum import DeviceType


class BaseScale(BaseDevice):
    device_dirname = "scales"
    device_type = DeviceType.SCALE

    def check_interfaces(self):
        driver_interfaces = providedBy(self._driver)
        if not IScale in driver_interfaces:
            raise TypeError("This driver doesn't implements a valid "
                            "interface")


def get_supported_scales():
    result = {}

    for brand, module_names in [(u'toledo', [u'PrixIII'])]:
        result[brand] = []
        for module_name in module_names:
            try:
                obj = namedAny("stoqdrivers.scales.%s.%s.%s"
                               % (brand, module_name, module_name))
            except AttributeError:
                raise ImportError("Can't find class %s for module %s"
                                  % (module_name, module_name))
            if not IScale.implementedBy(obj):
                raise TypeError("The driver %s %s doesn't implements a "
                                "valid interface"
                                % (brand, obj.model_name))
            result[brand].append(obj)
    return result



# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2005,2006 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Johan Dahlin     <jdahlin@async.com.br>
##              Henrique Romano  <henrique@async.com.br>
##

import logging

from serial import Serial, EIGHTBITS, PARITY_NONE, STOPBITS_ONE
from zope.interface import implements

from stoqdrivers.interfaces import ISerialPort
from stoqdrivers.exceptions import DriverError
from stoqdrivers.translation import stoqdrivers_gettext

_ = stoqdrivers_gettext

log = logging.getLogger('stoqdrivers.serial')


class VirtualPort:
    implements(ISerialPort)

    def getDSR(self):
        return True

    def setDTR(self, value):
        pass

    def write(self, data):
        pass

    def read(self, n_bytes=1):
        return ''


class SerialPort(Serial):
    implements(ISerialPort)

    def __init__(self, device, baudrate=9600):
        Serial.__init__(self, device)
        self.setDTR(True)
        self.flushInput()
        self.flushOutput()
        self.setBaudrate(baudrate)
        # WARNING: Never change these default options, some drivers are based
        # on this to work. Maybe we should change this and make all the driver
        # specify its options, but right now I think that's ok, since these
        # options are common to most of the drivers.
        self.setByteSize(EIGHTBITS)
        self.setParity(PARITY_NONE)
        self.setStopbits(STOPBITS_ONE)
        self.setTimeout(3)
        self.setWriteTimeout(0)


class SerialBase(object):

    # All commands will have this prefixed
    CMD_PREFIX = '\x1b'
    CMD_SUFFIX = ''

    # used by readline()
    EOL_DELIMIT = '\r'

    def __init__(self, port):
        self._port = port

    def set_port(self, port):
        self._port = port

    def get_port(self):
        return self._port

    def fileno(self):
        return self._port.fileno()

    def writeline(self, data):
        self.write(self.CMD_PREFIX + data + self.CMD_SUFFIX)
        return self.readline()

    def write(self, data):
        log.debug(">>> %r (%d bytes)" % (data, len(data)))
        self._port.write(data)

    def read(self, n_bytes):
        return self._port.read(n_bytes)

    def readline(self):
        out = ''
        a = 0
        retries = 10
        while True:
            if a > retries:
                raise DriverError(_("Timeout communicating with fiscal "
                                    "printer"))

            c = self._port.read(1)
            if not c:
                a += 1
                print 'take %s' % a
                continue
            a = 0
            if c == self.EOL_DELIMIT:
                log.debug('<<< %r' % out)
                return out
            out += c



# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2005-2007 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Evandro Vale Miquelito      <evandro@async.com.br>
##              Henrique Romano             <henrique@async.com.br>
##              Johan Dahlin                <jdahlin@async.com.br>
##
"""
StoqDrivers constants
"""

from stoqdrivers.enum import PaymentMethodType, TaxType, UnitType
from stoqdrivers.translation import stoqdrivers_gettext

_ = stoqdrivers_gettext

# TODO: Improve these descriptions
_constant_descriptions = {
    UnitType.WEIGHT: _(u"Weight unit"),
    UnitType.METERS: _(u"Meters unit"),
    UnitType.LITERS: _(u"Liters unit"),
    UnitType.EMPTY: _(u"Empty unit"),
    TaxType.ICMS: _(u"ICMS"),
    TaxType.SUBSTITUTION: _(u"Substitution"),
    TaxType.EXEMPTION: _(u"Exemption"),
    TaxType.NONE: _(u"No tax"),
    TaxType.SERVICE: _(u"Service tax"),
    PaymentMethodType.MONEY: _(u"Money Payment Method"),
    PaymentMethodType.CHECK: _(u"Check Payment Method"),
    PaymentMethodType.DEBIT_CARD: _(u"Debit card Payment Method"),
    PaymentMethodType.CREDIT_CARD: _(u"Credit card Payment Method"),
    PaymentMethodType.BILL: _(u"Bill Payment Method"),
    PaymentMethodType.FINANCIAL: _(u"Financial Payment Method"),
    PaymentMethodType.GIFT_CERTIFICATE: _(u"Gift certificate Payment Method"),
    PaymentMethodType.CUSTOM: _(u"Custom Payment Method"),
    PaymentMethodType.MULTIPLE: _(u"Multiple Payment Method"),
}


def describe_constant(constant_id):
    """ Given the constant identifier, return a short string describing it """
    global _constant_descriptions
    if not constant_id in _constant_descriptions:
        raise ValueError("The constant by id %r doesn't exists or there "
                         "is no description for it." % constant_id)
    return _constant_descriptions[constant_id]



# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2005 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Henrique Romano <henrique@async.com.br>
##
"""
Daruma FS2100 driver
"""

import operator
import logging
from decimal import Decimal
import time

from stoqdrivers.printers.base import BaseDriverConstants
from stoqdrivers.printers.daruma.FS345 import FS345, CMD_GET_TAX_CODES
from stoqdrivers.enum import UnitType, TaxType
from stoqdrivers.exceptions import DriverError

log = logging.getLogger('stoqdrivers.daruma')

ENQ = 05
ACK = 06
LF = 10
CR = 13
ESC = 27
FS = 28
GS = 29
FF = 255

CMD_ADD_ITEM = 201


class FS2100Constants(BaseDriverConstants):
    _constants = {
        UnitType.WEIGHT: 'Kg',
        UnitType.METERS: 'm ',
        UnitType.LITERS: 'Lt',
        UnitType.EMPTY: 'UN',
    }


class FS2100(FS345):
    model_name = "Daruma FS 2100"

    def __init__(self, port, consts=None):
        consts = consts or FS2100Constants
        FS345.__init__(self, port, consts)

    def setup(self):
        # Check printer to see how many decimal places it uses for price and
        # quantity
        data = self.send_new_command('R', 200, '139', ignore_error=True)
        self._decimals_qtd = Decimal('1e%s' % data[-2])
        self._decimals_price = Decimal('1e%s' % data[-1])

    def coupon_add_item(self, code, description, price, taxcode,
                        quantity=Decimal("1.0"), unit=UnitType.EMPTY,
                        discount=Decimal("0.0"),
                        surcharge=Decimal("0.0"), unit_desc=""):
        if surcharge:
            d = 3
            # Translating to human beings: surcharge * 100 or converting to cents
            E = int(surcharge * Decimal('1e2'))
        else:
            d = 1
            # Translating to human beings: discount * 100 or converting to cents
            E = int(discount * Decimal('1e2'))

        price = price
        quantity = quantity

        # The minimum size of the description, when working with one line for
        # description; if 0, write in multiple lines, if necessary.
        desc_size = 0

        if unit != UnitType.CUSTOM:
            unit = self._consts.get_value(unit)
        else:
            unit = "%2s" % unit_desc[:2]  # units must be 2 byte size strings

        # XXX: We need test correctly if the price's calcule is right (we
        # don't can do it right now since the manual isn't so clear).
        data = ('%02s'  # Tributary situation
                '%07d'  # Quantity
                '%08d'  # Unitary price
                '%d'   # 0=Discount(%) 1=Discount($) 2=Surcharge(%) 3=Surcharge($)
                '%011d'  # Discount/Surcharge value
                '%02d'  # Description size
                '%14s'  # Code
                '%3s'  # Unit of measure
                '%s'   # Product descriptio?!?
                '\xff'  # EOF
                % (taxcode, int(quantity * self._decimals_qtd),
                   int(price * self._decimals_price), d, E,
                   desc_size, code, unit, description[:233]))

        value = self.send_new_command('F', CMD_ADD_ITEM, data)
        return int(value[3:6])

    def send_command(self, command, extra=''):
        """As seen on ACBr code

        Nas Darumas com MFD, em algumas situações o ECF pode ficar
        temporariamente inoperantente, enquanto a compactação da MFD está
        sendo efetuada. Nessa situação, o ECF retorna os seguintes códigos
        de erro: 35 - Relogio Inoperante ou 99 (não documentado).

        Segundo o Suporte Técnico da Daruma, quando este problema ocorre,
        devemos aguardar até que ele consiga responder corretamente.

        Esta rotina irá tentar enviar o comando por 10 vezes. Caso ela
        recebe os erros 35 ou 99... ele aguarda 100 milisegundos e tenta um
        novo envio...
        """

        for t in range(10):
            try:
                retval = FS345.send_command(self, command, extra)
            except DriverError, e:
                if e.code == 99 or e.code == 35:
                    log.debug('FS2100 >>> Error 99. Sleeping for 0.1')
                    time.sleep(0.1)
                    continue

                raise
            break

        # After we have tried to send the command T times, we need to read
        # all the responses. This is really crap
        for i in range(t):
            reply = self._read_reply()
            log.debug('Ignoring reply: %r' % reply)

        return retval

    def _check_response(self, retcode, raw):
        # Compatible with the fs345/fs2100
        compatible_error = retcode[1:3]
        extended_error = retcode[3:6]
        warning_code = retcode[6:8]
        log.debug('FS2100 >>> Error %s - Extended: %s - Warning: %s' %
                  (compatible_error, extended_error, warning_code))
        if int(compatible_error):
            # Mimic FS345 error format
            error_code = ':E%s' % compatible_error
            self.handle_error(error_code, raw)

    def send_new_command(self, prefix, command, extra='', ignore_error=False):
        """ This method is used to send especific commands to model FS2100.
        Note that the main differences are the prefix (0x1c + 'F', since we
        will use a function of the FS2100 superset) and the checksum, that
        must be included in the end of all the functions of this superset.
        """
        data = chr(command) + extra

        log.debug('FS2100 >>> %r %d' % (data, len(data)))
        data = chr(FS) + prefix + data

        checksum = reduce(operator.xor, [ord(d) for d in data], 0)

        for t in range(10):
            self.write(data + chr(checksum))
            retval = self.readline()

            # See send_command for more details
            if retval == ':E99' or retval == ':E35':
                log.debug('FS2100 >>> Error 99. Sleeping for 0.1')
                time.sleep(0.1)
            else:
                break

        # After the CR, there is still one byte for the checksum
        retval_checksum = self.read(1)
        retval_checksum  # pyflakes

        if not ignore_error:
            self._check_response(retval, data)

        return retval[1:]

    def _get_compatibility_mode(self):
        return self.send_new_command('R', 200, '138')

    def till_add_cash(self, value):
        data = '%011d\xff' % (int(value * Decimal('1e2')))
        self.send_new_command('F', 236, data)

    def till_remove_cash(self, value):
        data = '%011d\xff' % (int(value * Decimal('1e2')))
        self.send_new_command('F', 227, data)

    def get_tax_constants(self):
        tax_codes = self.send_command(CMD_GET_TAX_CODES)[1:]
        lower_case = 'abcdefghijklmnop'

        constants = []
        for i in range(14):
            reg = tax_codes[i * 5]
            if reg in 'ABCDEFGHIJKLMNOP':
                tax_type = TaxType.CUSTOM
            elif reg in lower_case:
                tax_type = TaxType.SERVICE
            else:
                raise AssertionError(reg)
            value = tax_codes[i * 5 + 1:i * 5 + 5]
            if value == '////':
                continue
            constants.append((tax_type,
                              #'T' + reg.lower(),
                              '%02d' % (lower_case.index(reg.lower()) + 1),
                              Decimal(value.replace('.', '')) / 100))

        # These definitions can be found on Page 60
        constants.extend([
            (TaxType.SUBSTITUTION, '17', None),  # F1
            (TaxType.EXEMPTION, '19', None),  # I1
            (TaxType.NONE, '21', None),  # N1
        ])

        return constants



# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2012 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s): Stoq Team <stoq-devel@async.com.br>
##
"""
Epson FB II ECF driver
"""

import datetime
import logging
import struct
from decimal import Decimal

from kiwi.currency import currency
from kiwi.python import Settable
from zope.interface import implements

from stoqdrivers.serialbase import SerialBase
from stoqdrivers.interfaces import ICouponPrinter
from stoqdrivers.exceptions import (DriverError, PrinterError, CommandError,
                                    CommandParametersError, OutofPaperError,
                                    CancelItemError, AlmostOutofPaper,
                                    ItemAdditionError, CouponOpenError,
                                    CouponNotOpenError)
from stoqdrivers.enum import TaxType, UnitType
from stoqdrivers.printers.base import BaseDriverConstants
from stoqdrivers.translation import stoqdrivers_gettext

ACK = '\x06'
STX = '\x02'
ETX = '\x03'
ESC = '\x1b'
FLD = '\x1c'

# Special protocol chars that need to be escaped when communicating
SPECIAL_CHARS = [ESC, STX, ETX, FLD, '\x1a', '\x1d', '\x1e', '\x1f']


def escape(string):
    for i in SPECIAL_CHARS:
        string = string.replace(i, ESC + i)
    return string


def unescape(string):
    for i in SPECIAL_CHARS:
        string = string.replace(ESC + i, i)
    return string

FIRST_COMMAND_ID = 0x81
RETRIES_BEFORE_TIMEOUT = 5

# When cancel the last coupon. This values are used to define the coupon type.
FISCAL_COUPON = '1'
NON_FISCAL_COUPON = '24'

# Coupons status
CLOSED_COUPON = '0000'
OPENED_FISCAL_COUPON = '0001'
OPENED_NON_FISCAL_COUPON = '1000'

_ = stoqdrivers_gettext

log = logging.getLogger('stoqdrivers.epson')


class Reply(object):

    #
    #   Printer flags
    #

    error_codes = {
        '0101': (CommandError(_("Invalid command for current state."))),
        '0102': (CommandError(_("Invalid command for current document."))),
        '0203': (CommandError(_("Excess fields"))),
        '0204': (CommandError(_("Missing fields"))),
        '0205': (CommandParametersError(_("Field not optional."))),
        '0206': (CommandParametersError(_("Invalid alphanumeric field."))),
        '0207': (CommandParametersError(_("Invalid alphabetic field."))),
        '0208': (CommandParametersError(_("Invalid numeric field."))),
        '020E': (CommandParametersError(_("Fields with print invalid "
                                          "attributes."))),

        '0304': (OutofPaperError(_("Out of paper."))),
        '0305': (AlmostOutofPaper(_("Almost out of paper."))),

        '0801': (CommandError(_("Invalid command with closed "
                                "fiscal journey."))),

        '090C': (DriverError(_("Payment method not defined."))),
        '090F': (ItemAdditionError(_("Tax not found."))),
        '0910': (ItemAdditionError(_("Invalid tax."))),

        '0A12': (CancelItemError(_("It was not possible cancel the last "
                                   "fiscal coupon."))),
        '0A15': (PrinterError(_("Requires CDC cancellation."))),
        '0A16': (CancelItemError(_("Invalid item number in fiscal coupon"))),
        '0E0A': (PrinterError(_("Last non-fiscal coupon not found."))),
        '0E0B': (PrinterError(_("Payment method not found."))),
    }

    def __init__(self, string, command_id):
        checksum = string[-4:]
        self.string = string[:-4]

        log.debug('reply %s' % repr(string))
        cs = '%04X' % sum([ord(i) for i in self.string])
        if cs != checksum:
            raise DriverError('Erro de checksum')

        self.string = unescape(self.string)

        # Verifica header & tail
        assert self.pop() == STX, 'STX'
        frame_id = self.pop()
        if frame_id == '\x80':
            self.intermediate = True
            return

        self.intermediate = False
        assert frame_id == chr(command_id), ('command_id', command_id)
        assert self.string[-1] == ETX, 'ETX'

        # Retira statuses
        self.printer_status = struct.unpack('>H', self.pop(2))[0]
        assert self.pop() == FLD, 'FLD 1'
        self.fiscal_status = struct.unpack('>H', self.pop(2))[0]
        assert self.pop() == FLD, 'FLD 2'

        # reserved
        self.pop()

        r = self.pop(2)
        self.reply_status = '%02X%02X' % (ord(r[0]), ord(r[1]))

        assert self.pop() == FLD, 'FLD 3'
        # reserved
        self.pop()

        # Pega dados de retorno

        fields = self.string[:-1]

        # FIXME: Maybe we need to de-escape the string
        self.fields = fields.split(FLD)
        #for f in self.fields: print f

    def check_error(self):
        log.debug("reply_status %s" % self.reply_status)
        error_code = self.reply_status
        # Success, do nothing
        if error_code == '0000':
            return

        if error_code in self.error_codes:
            error = self.error_codes[error_code]
            error.code = int(error_code, 16)
            raise error

        raise DriverError(error="unhandled driver error",
                          code=int(error_code, 16))

    def pop(self, size=1):
        """Remove size bites from the begining of self.string and returns it
        """
        head = self.string[:size]
        self.string = self.string[size:]
        return head

    def check_printer_status(self):
        status = bin(self.printer_status)
        printer_status = status[2:]
        return printer_status

    def check_fiscal_status(self):
        status = bin(self.fiscal_status)
        fiscal_status = status[2:]
        return fiscal_status


class FBIIConstants(BaseDriverConstants):
    # The 'unit' field is mandatory in ECF.
    # If unit is EMPTY, set the value 'UN' to avoid errors.
    _constants = {
        UnitType.WEIGHT: 'Kg',
        UnitType.METERS: 'm',
        UnitType.LITERS: 'Lt',
        UnitType.EMPTY: 'UN',
    }


#
# The driver implementation
#

class FBII(SerialBase):

    implements(ICouponPrinter)

    supported = True
    identify_customer_at_end = True

    model_name = "Epson FBII"
    coupon_printer_charset = "ascii"

    def __init__(self, port, consts=None):
        SerialBase.__init__(self, port)
        self._consts = consts or FBIIConstants
        self._command_id = FIRST_COMMAND_ID - 1  # 0x80
        self._reset()

    def setup(self):
        # Get the decimal places from printer to use in price and quantity.
        data = self._send_command('0585')
        self._decimals_quantity = Decimal('1e' + data.fields[0])
        self._decimals_price = Decimal('1e' + data.fields[1])

    def _reset(self):
        self._customer_name = ''
        self._customer_document = ''
        self._customer_address = ''

    #
    # Helper methods
    #

    def _get_next_command_id(self):
        if self._command_id == 255:
            self._command_id = FIRST_COMMAND_ID - 1

        self._command_id += 1
        return chr(self._command_id)

    def _get_package(self, command, extension, args=None):
        # First, convert command and extention from string representation to hex
        command = '%s%s' % (chr(int(command[:2], 16)), chr(int(command[2:], 16)))
        extension = '%s%s' % (chr(int(extension[:2], 16)), chr(int(extension[2:], 16)))

        frame = escape(command) + FLD + escape(extension)

        if args:
            for i in args:
                frame += FLD + escape(i)

        command_id = self._get_next_command_id()
        package = STX + command_id + frame + ETX

        checksum = sum([ord(d) for d in package])
        return package + '%04X' % checksum

    def _read_reply(self):
        reply = ''
        timeouts = 0

        while True:
            if timeouts > RETRIES_BEFORE_TIMEOUT:
                raise DriverError(_("Timeout communicating with fiscal "
                                    "printer"))

            c = self.read(1)
            if len(c) != 1:
                timeouts += 1
                continue

            # STX is always the first char in the reply. Ignore garbage
            # until STX is received.
            if len(reply) == 0 and c != STX:
                log.info('ignoring garbage in reply: %r' % c)
                continue

            reply += c

            if c == ETX and reply[-2] != ESC:
                break

        reply += self.read(4)
        log.debug("<<< %s" % repr(reply))

        return Reply(reply, self._command_id)

    def _send_command(self, command, extension='0000', *args):
        cmd = self._get_package(command, extension, args)
        #log.debug("> %s" % repr(cmd))
        self.write(cmd)

        # Printer should reply with an ACK imediataly
        ack = self.read(1)
        if not ack:
            raise DriverError(_("Timeout communicating with fiscal "
                                "printer"))
        assert ack == ACK, repr(ack)

        reply = self._read_reply()

        # Keep reading while printer sends intermediate replies.
        while reply.intermediate:
            log.debug("intermediate")
            reply = self._read_reply()

        # send our ACK
        self.write(ACK)

        reply.check_error()
        return reply

    def _parse_price(self, value):
        # Valor retirado da ECF (string) convertido para decimal.
        return currency(value) / Decimal('1e2')

    #
    # ICouponPrinter implementation
    #

    # Printer Quering

    def query_status(self):
        cmd = self._get_package('0001', '0000')
        return cmd

    def status_reply_complete(self, reply):
        complete = len(reply) == 18
        if complete:
            self.write(ACK)
        return complete

    #
    #   Till Operations
    #

    def summarize(self):
        self._send_command('0802')

    def open_till(self):
        self._send_command('0805')

    def has_pending_reduce(self):
        # Dia encerrado e NÃO reduzido.
        pending_reduce_z = 4

        # Estado da jornada fiscal.
        reply = self._send_command('0810')
        return int(reply.fields[0]) == pending_reduce_z

    def close_till(self, previous_day=False):
        # Nesse momento é possível realizar ajustes no horário da ECF (como
        # entrar e sair do horário de verão. Podemos estudar implementar isso no
        # stoq.
        self._send_command('0801', '0000', '', '')

    def _add_voucher(self, type, value):
        # Abre comprovante não fiscal
        self._send_command('0E01', '0000', '')

        # Adiciona item
        value = int(value * Decimal('1e2'))
        self._send_command('0E15', '0000', type, str(value))

        # Fecha
        self._send_command('0E06')

    def till_add_cash(self, value):
        self._add_voucher('02', value)

    def till_remove_cash(self, value):
        self._add_voucher('01', value)

    def till_read_memory(self, start, end):
        self._send_command('0910', '0001', '', '', start.strftime('%d%m%Y'),
                           end.strftime('%d%m%Y'))

    def till_read_memory_by_reductions(self, start, end):
        self._send_command('0910', '0000', str(start), str(end), '', '')

    def get_sintegra(self):
        # Informações da jornada fiscal.
        reply = self._send_command('080A')
        # Data de abertura.
        date = reply.fields[0]
        d, m, y = map(int, [date[:2], date[2:4], date[4:8]])
        opening_date = datetime.date(y, m, d)
        # COO inicial.
        coupon_start = int(reply.fields[4])

        #Número sequencial da ECF.
        fiscal_data = self._send_command('0507')
        serial_id = int(fiscal_data.fields[8])

        # Obter totalizadores.
        ecf_totals = self._send_command('0906')
        # Totalizador venda bruta diária.
        value = ecf_totals.fields[1]
        period_total = currency(value) / Decimal('1e2')
        # Total Geral.
        geral = ecf_totals.fields[0]
        total = currency(geral) / Decimal('1e2')

        data = Settable(
            opening_date=opening_date,
            serial=self.get_serial(),
            serial_id=serial_id,
            coupon_start=coupon_start,
            coupon_end=self.get_coo(),
            cro=self.get_cro(),
            crz=self.get_crz(),
            coo=self.get_coo(),
            period_total=period_total,
            total=total,
            taxes=self._get_taxes(),
        )
        return data

    def get_payment_receipt_identifier(self, method_name):
        # We don't need a identifier. Just return None.
        return None

    def payment_receipt_open(self, identifier, coo, method, value):
        # FIXME: Verificar possibilidade de se implementar a utilização de
        # quantidade de parcelas.
        value = int(value * Decimal('1e2'))
        self._send_command('0E30', '0000', method, str(value), '1', str(coo))

    def payment_receipt_print(self, text):
        self._print_report_text(text)

    def payment_receipt_close(self):
        self._send_command('0E06')

    #
    # Coupon related commands
    #

    def coupon_identify_customer(self, customer, address, document):
        self._customer_name = customer
        self._customer_document = document
        self._customer_address = address

    def coupon_is_customer_identified(self):
        return len(self._customer_document) > 0

    def has_open_coupon(self):
        checked_coupon = self._get_coupon_status()
        return checked_coupon != CLOSED_COUPON

    def coupon_open(self):
        coupon_status = self._get_coupon_status()
        if coupon_status == OPENED_FISCAL_COUPON:
            raise CouponOpenError(_("Coupon already open"))
        self._send_command('0A01', '0000', '', '')

    def _cancel_fiscal_coupon(self):
        self._send_command('0A18', '0008', '1')

    def _cancel_non_fiscal_coupon(self):
        self._send_command('0E18', '0008', '1')

    def coupon_cancel(self):
        checked_coupon = self._get_coupon_status()
        if checked_coupon == OPENED_FISCAL_COUPON:
            self._cancel_fiscal_coupon()
        elif checked_coupon == OPENED_NON_FISCAL_COUPON:
            self._cancel_non_fiscal_coupon()

    def coupon_add_item(self, code, description, price, taxcode,
                        quantity=Decimal("1.0"), unit=UnitType.EMPTY,
                        discount=Decimal("0.0"), markup=Decimal("0.0"),
                        unit_desc=""):
        coupon_status = self._get_coupon_status()
        if coupon_status != OPENED_FISCAL_COUPON:
            raise CouponNotOpenError(_("Coupon is not open"))

        if unit == UnitType.CUSTOM:
            unit = unit_desc
        else:
            unit = self._consts.get_value(unit)
        qtd = int(quantity * self._decimals_quantity)
        value = int(price * self._decimals_price)
        st = taxcode
        # Register the sale item
        reply = self._send_command('0A02', '0000', code, description,
                                   str(qtd), unit, str(value), st)
        id = reply.fields[0]
        if discount:
            self.apply_discount(id, discount)
        elif markup:
            self.apply_markup(id, markup)
        return int(id)

    def apply_discount(self, id, discount):
        value = int(discount * Decimal('1e2'))
        # Register the discount on the item
        self._send_command('0A04', '0004', str(value))

    def apply_markup(self, id, markup):
        value = int(markup * Decimal('1e2'))
        # Register the markup on the item
        self._send_command('0A04', '0005', str(value))

    def coupon_cancel_item(self, item_id):
        self._send_command('0A18', '0004', str(item_id))

    def cancel_last_coupon(self):
        # Informações sobre último documento.
        reply = self._send_command('0908')
        last_coupon = reply.fields[0]
        if last_coupon == FISCAL_COUPON:
            self._cancel_fiscal_coupon()
        elif last_coupon == NON_FISCAL_COUPON:
            self._cancel_non_fiscal_coupon()
        else:
            raise DriverError(_("Attempt to cancel after emission of another "
                                "DOC"))

    def coupon_totalize(self, discount=currency(0), markup=currency(0),
                        taxcode=TaxType.NONE):
        if discount:
            extension = '0006'
            value = discount
        else:
            extension = '0007'
            value = markup

        # Podemos somente mandar o comando de desconto/acrescimo se realmente
        # tivermos um valor.
        if value:
            value = int(value * Decimal('1e2'))
            reply = self._send_command('0A04', extension, str(value))
            subtotal = reply.fields[0]
        else:
            # Ainda precisamos pegar o subtotal, para retornar.
            reply = self._send_command('0A03')
            subtotal = reply.fields[0]
        return currency(subtotal) / Decimal('1e2')

    def coupon_add_payment(self, payment_method, value, description=u""):
        desc = description[:40]
        value = int(value * Decimal('1e2'))
        reply = self._send_command('0A05', '0000', payment_method, str(value),
                                   desc, '')
        # Return the still missing value
        return currency(reply.fields[0]) / Decimal('1e2')

    def coupon_close(self, message=""):
        if self._customer_document:
            self._send_command('0A20', '0002', self._customer_document[:20],
                               self._customer_name[:30],
                               self._customer_address[:40],
                               self._customer_address[41:79])
            self._reset()

        if message:
            self._print_promotional_message(message)

        reply = self._send_command('0A06', '0001')
        self._send_command('0702', '0000')
        return int(reply.fields[0])

    def gerencial_report_open(self):
        return self._send_command('0E01', '0004', '1')

    def gerencial_report_print(self, text):
        self._print_report_text(text)

    def gerencial_report_close(self):
        self._send_command('0E06')

    def _print_report_text(self, text):
        """ Format the text according to limit of collumns number.
        """

        # Number of characters per line must be less than 56.
        for line in text.split('\n'):
            if line == '':
                self._print_line(line)
            for pos in range(0, len(line), 56):
                data = line[pos:pos + 56]
                self._print_line(data)

    def _print_line(self, line):
        """
        As seen on ACBr code.
        Remove char '\r'. When the ECF receives the character \r,
        the error, 020E(invalid attribute) is emitted.
        """
        line = line.strip('\r')
        self._send_command('0E02', '0000', line)

    def _print_promotional_message(self, message):
        msg = []
        # Epson accepts only 56 chars per line. Lets truncate each line
        for line in message.split('\n'):
            msg.append(line[:56])

        # We must send 8 lines to fiscal printer, even if they are empty.
        while len(msg) < 8:
            msg.append('')
        self._send_command('0A22', '0000', *msg[:8])

    #
    #   General information
    #

    def get_serial(self):
        reply = self._get_ecf_details()
        return reply.fields[0]

    def get_tax_constants(self):
        reply = self._send_command('0542')
        constants = []

        fields = reply.fields
        if len(fields) > 2:
            for i in range(0, len(fields), 3):
                name, value = fields[i], fields[i + 1]

                if name.startswith('T'):
                    type = TaxType.CUSTOM
                elif name.startswith('S'):
                    type = TaxType.SERVICE
                else:
                    assert False, (name, value)

                constants.append((type, name, Decimal(value) / 100))

        constants.extend([
            (TaxType.SUBSTITUTION, 'F', None),
            (TaxType.EXEMPTION, 'I', None),
            (TaxType.NONE, 'N', None),
        ])

        return constants

    def get_payment_constants(self):
        methods = []

        for i in range(20):
            try:
                reply = self._send_command('050D', '0000', '%d' % (i + 1))
            except DriverError, e:
                if e.code == 0x090C:  # Tipo de pagamento não definido
                    continue
                else:
                    raise
            name, vinculado = reply.fields
            methods.append(('%d' % (i + 1), name.strip()))

        return methods

    def get_capabilities(self):
        from stoqdrivers.printers.capabilities import Capability
        return dict(
            item_code=Capability(max_len=14),
            item_description=Capability(max_len=233),
            customer_id=Capability(max_len=28),
            customer_name=Capability(max_len=30),
            customer_address=Capability(max_len=79),
        )

    def get_constants(self):
        return self._consts

    def _get_ecf_details(self):
        return self._send_command('0402')

    def get_firmware_version(self):
        reply = self._get_ecf_details()
        return reply.fields[5]

    def get_coo(self):
        reply = self._send_command('0907')
        return int(reply.fields[0])

    def get_gnf(self):
        reply = self._send_command('0907')
        return int(reply.fields[3])

    def get_ccf(self):
        reply = self._send_command('0907')
        return int(reply.fields[7])

    def get_cro(self):
        reply = self._send_command('0907')
        return int(reply.fields[2])

    def get_crz(self):
        reply = self._send_command('0907')
        return int(reply.fields[1])

    def _get_taxes(self):
        reply = self._send_command('0542')
        tax_codes = reply.fields

        taxes = []
        for i in range(20):
            reg = tax_codes[i * 3:i * 3 + 3]
            if not reg:
                continue

            name = reg[0]
            if name.startswith('T'):
                tax_type = 'ICMS'
            else:
                tax_type = 'ISS'

            sold = reg[2]
            value = reg[1]
            taxes.append((value, self._parse_price(sold), tax_type))
        tax_totals = self._send_command('0906')

        taxes.append(('I', self._parse_price(tax_totals.fields[16]), 'ICMS'))
        taxes.append(('F', self._parse_price(tax_totals.fields[15]), 'ICMS'))
        taxes.append(('N', self._parse_price(tax_totals.fields[17]), 'ICMS'))
        taxes.append(('DESC', self._parse_price(tax_totals.fields[3]), 'ICMS'))
        taxes.append(('CANC', self._parse_price(tax_totals.fields[2]), 'ICMS'))

        return taxes

    #
    # Status
    #

    def _get_fiscal_status(self):
        reply = self._send_command('0001')
        return reply.check_fiscal_status()

    def _get_printer_status(self):
        reply = self._send_command('0001')
        return reply.check_printer_status()

    def _get_coupon_status(self):
        # Last four digits of fiscal status, show coupon status
        return self._get_fiscal_status()[12:16]

    #
    #   Printer configuration
    #

    def _define_payment_method(self, id, name, vinculated=False):
        if vinculated:
            extension = '0001'
        else:
            extension = '0000'

        id = '%02d' % id
        self._send_command('050C', extension, id, name)

    def _define_tax_code(self, value, service=False):
        if service:
            extension = '0001'
        else:
            extension = '0000'

        self._send_command('0540', extension, value)

    def _setup_constants(self):
        self._define_payment_method(2, 'Cheque')
        self._define_payment_method(3, 'Boleto')
        self._define_payment_method(4, 'Cartao credito', vinculated=True)
        self._define_payment_method(5, 'Cartao debito', vinculated=True)
        self._define_payment_method(6, 'Financeira')
        self._define_payment_method(7, 'Vale compra')

        self._define_tax_code("1700")
        self._define_tax_code("1200")
        self._define_tax_code("2500")
        self._define_tax_code("0800")
        self._define_tax_code("0500")
        self._define_tax_code("0300", service=True)
        self._define_tax_code("0900", service=True)


if __name__ == '__main__':
    from stoqdrivers.serialbase import SerialPort
    #port = Serial('/dev/pts/5')
    port = SerialPort('/dev/ttyUSB0')
    p = FBII(port)

#    p._send_command('0754', '0000') Desativar corte do papel

#    p._setup_constants()
#    constants = p.get_tax_constants()
#    for i in constants:
#        print i

#    constants = p.get_payment_constants()
#    for i in constants:
#        print i



#!/usr/bin/env python

# Setup file for StoqDrivers
# Code by Async Open Source <http://www.async.com.br>

from kiwi.dist import setup, listpackages, listfiles

from stoqdrivers import __version__


with open('requirements.txt') as f:
    install_requires = [l.strip() for l in f.readlines() if
                        l.strip() and not l.startswith('#')]

setup(
    name="stoqdrivers",
    version=".".join(map(str, __version__)),
    author="Async Open Source",
    author_email="stoq-devel@async.com.br",
    description="Python fiscal printer (ECF) drivers",
    long_description=("This package provices device drivers "
                      "for fiscal printers, ECF (Emissor de Coupon Fiscal) "
                      "written in Python. Supports printers from Bematech, "
                      "Daruma, Dataregis, Perto, Sweda and the generic "
                      "FiscNET protocol."),
    url="http://www.stoq.com.br",
    license="GNU LGPL 2.1 (see COPYING)",
    packages=listpackages('stoqdrivers'),
    data_files=[("$datadir/conf", listfiles("stoqdrivers/conf", "*.ini"))],
    install_requires=install_requires,
)



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Copyright (C) 2015 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., or visit: http://www.gnu.org/.
##
## Author(s): Stoq Team <stoq-devel@async.com.br>
##

from stoqdrivers.printers.base import BasePrinter


class NonFiscalPrinter(BasePrinter):
    def __init__(self, brand=None, model=None, device=None, config_file=None,
                 *args, **kwargs):
        BasePrinter.__init__(self, brand, model, device, config_file, *args,
                             **kwargs)

    def centralize(self):
        self._driver.centralize()

    def descentralize(self):
        self._driver.descentralize()

    def set_bold(self):
        self._driver.set_bold()

    def unset_bold(self):
        self._driver.unset_bold()

    def print_line(self, data):
        self._driver.print_line(data)

    def print_inline(self, data):
        self._driver.print_inline(data)

    def print_barcode(self, barcode):
        self._driver.print_barcode(barcode)

    def print_qrcode(self, code):
        self._driver.print_qrcode(code)



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Copyright (C) 2007 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s): Johan Dahlin              <jdahlin@async.com.br>
##
#
# Documentation references:
#
#   http://en.wikipedia.org/wiki/ESC/P
#   http://www.epson.co.uk/support/manuals/pdf/ESCP/Part_1.pdf
#   http://www.epson.co.uk/support/manuals/pdf/ESCP/Part_2.pdf
#
""" Driver for EPSON Esc/P and Esc/P2 printers. """

import struct

ESC = '\x1b'

CMD_INIT = '@'
CMD_PRINT_QUALITY = 'x'
CMD_PROPORTIONAL = 'p'
CMD_FORM_FEED = '\xff'
CMD_EJECT = '\x19'

QUALITY_DRAFT = '0'
QUALITY_LQ = '1'
QUALITY_NLQ = '1'


class EscPPrinter(object):
    def __init__(self, device):
        self.device = device
        self.fp = open(device, 'w')

        self._command(CMD_INIT)

    def _command(self, command, *args):
        chars = command
        for arg in args:
            if arg is True:
                v = '1'
            elif arg is False:
                v = '0'
            else:
                v = arg

            chars += v
        cmd = '%s%s' % (ESC, chars)
        self.send(cmd)

    def send(self, data):
        self.fp.write(data)
        self.fp.flush()

    def set_draft_mode(self):
        self._command(CMD_PRINT_QUALITY, QUALITY_DRAFT)

    def set_proportional(self, proportional):
        self._command(CMD_PROPORTIONAL, proportional)

    def done(self):
        self._command(CMD_INIT)

    def form_feed(self):
        self._command(CMD_FORM_FEED)

    def set_vertical_position(self, position):
        args = struct.pack('b', position)
        self._command('J', *args)


def test():
    printer = EscPPrinter('/dev/lp0')
    printer.send(
        'Lorem ipsum dolor sit amet, consectetuer adipiscing elit. '
        'Ut a velit sit amet nisl hendrerit lacinia. Nunc eleifend '
        'cursus risus. Vivamus libero libero, dignissim ut, pulvinar id, '
        'blandit a, leo amet.\n'.upper())

    printer.done()

if __name__ == '__main__':
    test()



# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2005-2007 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Cleber Rodrigues      <cleber@globalred.com.br>
##              Henrique Romano       <henrique@async.com.br>
##              Johan Dahlin          <jdahlin@async.com.br>
##              Ronaldo Maia          <romaia@async.com.br>
##
"""
Bematech MP25 driver
"""

import datetime
from decimal import Decimal
import logging
import struct

from kiwi.currency import currency
from kiwi.python import Settable
from zope.interface import implements

from stoqdrivers.serialbase import SerialBase
from stoqdrivers.exceptions import (DriverError, OutofPaperError, PrinterError,
                                    CommandError, CouponOpenError,
                                    HardwareFailure,
                                    PrinterOfflineError, PaymentAdditionError,
                                    ItemAdditionError, CancelItemError,
                                    CouponTotalizeError, CouponNotOpenError)
from stoqdrivers.interfaces import ICouponPrinter
from stoqdrivers.printers.capabilities import Capability
from stoqdrivers.printers.base import BaseDriverConstants
from stoqdrivers.enum import TaxType, UnitType
from stoqdrivers.translation import stoqdrivers_gettext

_ = stoqdrivers_gettext

log = logging.getLogger('stoqdrivers.bematech.MP25')

CASH_IN_TYPE = "SU"
CASH_OUT_TYPE = "SA"

# The comment after the commands are comparing the MP25 command with MP20.

CMD_COUPON_OPEN = 0  # MP25 aceita 2 parametros opcionais a mais
CMD_CLOSE_TILL = 5
CMD_REDUCE_Z = 5
CMD_READ_X = 6
CMD_ADD_TAX = 7
CMD_READ_MEMORY = 8
CMD_COUPON_CANCEL = 14  # MP25 aceita 3 parametros opicionais
CMD_STATUS = 19
CMD_ADD_VOUCHER = 25
CMD_READ_TAXCODES = 26
CMD_READ_TOTALIZERS = 27
CMD_GET_COUPON_SUBTOTAL = 29
CMD_GET_COUPON_NUMBER = 30
CMD_CANCEL_ITEM = 31
CMD_COUPON_TOTALIZE = 32  # MP 20 Nao permite desconto simultaneo
CMD_COUPON_CLOSE = 34
CMD_READ_REGISTER = 35  # Registradores diferentes
CMD_ADD_ITEM = 63
CMD_PAYMENT_RECEIPT_OPEN = 66  # MP20 suporta somente 3 primeiro parametros
CMD_PAYMENT_RECEIPT_PRINT = 67
CMD_PAYMENT_RECEIPT_CLOSE = 21  # Parece o mesmo, mas com nomes diferentes na mp20 e 25
CMD_GERENCIAL_REPORT_PRINT = 20
CMD_GERENCIAL_REPORT_CLOSE = 21

CMD_PROGRAM_PAYMENT_METHOD = 71
CMD_ADD_PAYMENT = 72

# Be carefull with these commands, They are only available for MP25 (not mp20)
CMD_CANCEL_LAST = 81
CMD_PAYMENT_RECEIPT_PRINT_DUPLICATE = 91

NAK = 21
ACK = 6
STX = 2

RETRIES_BEFORE_TIMEOUT = 5
CHARS_LIMIT = 492
ALLOW_CANCEL_FISCAL_COUPON = 32


# Page 51
class MP25Registers(object):
    TOTAL = 3
    TOTAL_CANCELATIONS = 4
    TOTAL_DISCOUNT = 5
    COO = 6
    GNF = 7
    NUMBER_REDUCTIONS_Z = 9
    CRO = 10
    LAST_ITEM_ID = 12
    NUMBER_TILL = 14
    FISCAL_FLAGS = 17
    EMISSION_DATE = 23
    TRUNC_FLAG = 28
    TOTALIZERS = 29
    PAYMENT_METHODS = 32
    SERIAL = 40
    FIRMWARE = 41
    CCF = 55

    # (size, bcd)
    formats = {
        TOTAL: ('9s', True),
        TOTAL_CANCELATIONS: ('7s', True),
        TOTAL_DISCOUNT: ('7s', True),
        COO: ('3s', True),
        GNF: ('3s', True),
        NUMBER_REDUCTIONS_Z: ('2s', True),
        CRO: ('2s', True),
        LAST_ITEM_ID: ('2s', True),
        NUMBER_TILL: ('2s', True),
        FISCAL_FLAGS: ('1s', False),
        EMISSION_DATE: ('6s', False),
        TRUNC_FLAG: ('1s', False),
        TOTALIZERS: ('2s', False),
        #  1 + (52 * 16) + (52 * 10) + (52 * 10) + (52 * 1)
        #  1 + 832 + 520 + 520 + 52: 1925
        PAYMENT_METHODS: ('b832s520s520s52s', False),
        SERIAL: ('20s', False),
        FIRMWARE: ('3s', True),
        CCF: ('3s', True),
    }


class MP25Constants(BaseDriverConstants):
    _constants = {
        UnitType.WEIGHT: 'Kg',
        UnitType.METERS: 'm ',
        UnitType.LITERS: 'Lt',
        UnitType.EMPTY: '  ',
    }


class MP25Status(object):
    PENDING_REDUCE_Z = 66

    st1_codes = {
        128: (OutofPaperError(_("Printer is out of paper"))),
        # 64: (AlmostOutofPaper(_("Printer almost out of paper"))),
        32: (PrinterError(_("Printer clock error"))),
        16: (PrinterError(_("Printer in error state"))),
        8: (CommandError(_("First data value in CMD is not ESC (1BH)"))),
        4: (CommandError(_("Nonexistent command"))),
        # 2: (CouponOpenError(_("Printer has a coupon currently open"))),
        1: (CommandError(_("Invalid number of parameters")))}

    st2_codes = {
        128: (CommandError(_("Invalid CMD parameter"))),
        64: (HardwareFailure(_("Fiscal memory is full"))),
        32: (HardwareFailure(_("Error in CMOS memory"))),
        16: (PrinterError(_("Given tax is not programmed on the printer"))),
        8: (DriverError(_("No available tax slot"))),
        4: (CancelItemError(_("The item wasn't added in the coupon or can't "
                              "be cancelled"))),

        # 2: (PrinterError(_("Owner data (CGC/IE) not programmed on the printer"))),
        # FIXME: This shouldn't be commented. But it will break the tests.
        # Need to update the tests for all bematech printers
        #1: (CommandError(_("Command not executed")))
    }

    st3_codes = {
        # 7: (CouponOpenError(_("Coupon already Open"))),
        # 8: (CouponNotOpenError(_("Coupon is closed"))),
        13: (PrinterOfflineError(_("Printer is offline"))),
        16: (DriverError(_("Surcharge or discount greater than coupon total"
                           "value"))),
        17: (DriverError(_("Coupon with no items"))),
        20: (PaymentAdditionError(_("Payment method not recognized"))),
        22: (PaymentAdditionError(_("Isn't possible add more payments since"
                                    "the coupon total value already was "
                                    "reached"))),
        23: (DriverError(_("Coupon isn't totalized yet"))),
        43: (CouponNotOpenError(_("Printer not initialized"))),
        45: (PrinterError(_("Printer without serial number"))),
        52: (DriverError(_("Invalid start date"))),
        53: (DriverError(_("Invalid final date"))),
        85: (DriverError(_("Sale with null value"))),
        91: (ItemAdditionError(_("Surcharge or discount greater than item"
                                 "value"))),
        100: (DriverError(_("Invalid date"))),
        115: (CancelItemError(_("Item doesn't exists or already was cancelled"))),
        118: (DriverError(_("Surcharge greater than item value"))),
        119: (DriverError(_("Discount greater than item value"))),
        129: (CouponOpenError(_("Invalid month"))),
        169: (CouponTotalizeError(_("Coupon already totalized"))),
        170: (PaymentAdditionError(_("Coupon not totalized yet"))),
        171: (DriverError(_("Surcharge on subtotal already effected"))),
        172: (DriverError(_("Discount on subtotal already effected"))),
        176: (DriverError(_("Invalid date")))}

    def __init__(self, reply):
        self.st1, self.st2, self.st3 = reply[-3:]

    @property
    def open(self):
        return self.st1 & 2

    def _check_error_in_dict(self, error_codes, value):
        for key in error_codes:
            if key & value:
                raise error_codes[key]

    def check_error(self):
        log.debug("status: st1=%s st2=%s st3=%s" %
                  (self.st1, self.st2, self.st3))

        if self.st1 != 0:
            self._check_error_in_dict(self.st1_codes, self.st1)

        if self.st2 != 0:
            self._check_error_in_dict(self.st2_codes, self.st2)

            # first bit means not executed, look in st3 for more
            if self.st2 & 1 and self.st3:
                if self.st3 in self.st3_codes:
                    raise self.st3_codes[self.st3]


#
# Helper functions
#

def bcd2dec(data):
    return int(''.join(['%02x' % ord(i) for i in data]))


def bcd2hex(data):
    return ''.join(['%02x' % ord(i) for i in data])


def dec2bcd(dec):
    return chr(dec % 10 + (dec / 10) * 16)


def dec2bin(n, trim=-1):
    a = ""
    while n > 0:
        if n % 2 == 0:
            a = "0" + a
        else:
            a = "1" + a
        n /= 2

    if trim != -1:
        if len(a) < trim:
            a = ("0" * (trim - len(a))) + a
    return a

#
# Driver implementation
#


class MP25(SerialBase):
    implements(ICouponPrinter)
    CMD_PROTO = 0x1c

    supported = True
    model_name = "Bematech MP25 FI"
    coupon_printer_charset = "cp850"
    supports_duplicate_receipt = True
    identify_customer_at_end = False
    registers = MP25Registers
    reply_format = '<b%sbbH'
    status_size = 3

    EOL_DELIMIT = '\n'

    def __init__(self, port, consts=None):
        self._consts = consts or MP25Constants
        port.setTimeout(2)
        port.setWriteTimeout(5)
        SerialBase.__init__(self, port)
        # XXX: Seems that Bematech doesn't contains any variable with the
        # coupon remainder value, so I need to manage it by myself.
        self.remainder_value = Decimal("0.00")
        self._reset()

    def _reset(self):
        self._customer_name = ''
        self._customer_document = ''
        self._customer_address = ''

    def _create_packet(self, command):
        """
        Create a 'pre-package' (command + params, basically) and involves
        it around STX, NB and CS::
           1     2           n           2
        +-----+------+-----------------+----+
        | STX |  NB  | 0x1C + command  | CS |
        +-----+------+-----------------+----+

        Where:

        STX: 'Transmission Start' indicator byte
        NB: 2 bytes, big endian length of command + CS (2 bytes)
        CS: 2 bytes, big endian checksum for command
        """

        command = chr(self.CMD_PROTO) + command
        return struct.pack('<bH%dsH' % len(command),
                           STX,
                           len(command) + 2,
                           command,
                           sum([ord(i) for i in command]))

    def _read_reply(self, size):
        a = 0
        data = ''
        while True:
            if a > RETRIES_BEFORE_TIMEOUT:
                raise DriverError(_("Timeout communicating with fiscal "
                                    "printer"))

            a += 1
            reply = self.read(size)
            if reply is None:
                continue

            data += reply
            if len(data) < size:
                continue

            log.debug("<<< %r (%d bytes)" % (data, len(data)))
            return data

    def _check_error(self, retval=None):
        status = self.get_status(retval)
        status.check_error()

    def _send_command(self, command, *args, **kwargs):
        fmt = ''
        if 'response' in kwargs:
            fmt = kwargs.pop('response')

        raw = False
        if 'raw' in kwargs:
            raw = kwargs.pop('raw')

        if kwargs:
            raise TypeError("Invalid kwargs: %r" % (kwargs,))

        cmd = chr(command)
        for arg in args:
            if isinstance(arg, int):
                cmd += chr(arg)
            elif isinstance(arg, str):
                cmd += arg
            else:
                raise NotImplementedError(type(arg))

        data = self._create_packet(cmd)
        self.write(data)

        format = self.reply_format % fmt
        reply = self._read_reply(struct.calcsize(format))
        retval = struct.unpack(format, reply)

        if raw:
            return retval

        self._check_error(retval)

        response = retval[1:-self.status_size]
        if len(response) == 1:
            response = response[0]
        return response

    def _read_register(self, reg):
        try:
            fmt, bcd = self.registers.formats[reg]
        except KeyError:
            raise NotImplementedError(reg)

        value = self._send_command(CMD_READ_REGISTER, reg, response=fmt)
        if bcd:
            value = bcd2dec(value)
        return value

    #
    # Helper methods
    #

    def _get_coupon_subtotal(self):
        subtotal = self._send_command(CMD_GET_COUPON_SUBTOTAL, response='7s')
        if subtotal:
            return Decimal(bcd2dec(subtotal)) / Decimal("1e2")
        # Busted subtotal
        return Decimal("0.0")

    def _get_last_item_id(self):
        return self._read_register(self.registers.LAST_ITEM_ID)

    def _get_coupon_number(self):
        coupon_number = self._send_command(CMD_GET_COUPON_NUMBER, response='3s')
        return bcd2dec(coupon_number)

    def get_status(self, val=None):
        if val is None:
            val = self._send_command(CMD_STATUS, raw=True)

        return MP25Status(val)

    def _add_voucher(self, type, value):
        assert len(type) == 2

        status = self.get_status()
        if status.open:
            self._send_command(CMD_COUPON_CANCEL)

        self._send_command(CMD_ADD_VOUCHER, type, "%014d" % int(value * Decimal('1e2')))

    def _setup_constants(self):
        return
        # Do one at a time, if you need it.
        #self._send_command(CMD_PROGRAM_PAYMENT_METHOD,
        #                         '%-16s1' % 'Cartao Credito', raw=True)
        #self._send_command(CMD_PROGRAM_PAYMENT_METHOD,
        #                         '%-16s1' % 'Cartao Debito', raw=True)
        #self._send_command(CMD_PROGRAM_PAYMENT_METHOD,
        #                         '%-16s0' % 'Cheque', raw=True)

        #self._send_command(CMD_ADD_TAX, '25000', raw=True) # ICMS
        #self._send_command(CMD_ADD_TAX, '17000', raw=True) # ICMS
        #self._send_command(CMD_ADD_TAX, '12000', raw=True) # ICMS
        #self._send_command(CMD_ADD_TAX, '08000', raw=True) # ICMS
        #self._send_command(CMD_ADD_TAX, '05000', raw=True) # ICMS
        #self._send_command(CMD_ADD_TAX, '03001', raw=True) # ISS

    #
    # This implements the ICouponPrinter Interface
    #

    def setup(self):
        pass

    def summarize(self):
        """ Prints a summary of all sales of the day """
        self._send_command(CMD_READ_X)

    def open_till(self):
        self.summarize()

    def has_pending_reduce(self):
        status = self.get_status()
        return status.st3 == MP25Status.PENDING_REDUCE_Z

    def close_till(self, previous_day=False):
        """ Close the till for the day, no other actions can be done after this
        is called.
        """
        # a impressao da Leitura X precisa ser feita diariamente. Ao utilizar o
        # parâmetro previous_day a Leitura X só será impressa quando o caixa
        # não é fechado no mesmo dia
        self._send_command(CMD_REDUCE_Z)

    def till_add_cash(self, value):
        self._add_voucher(CASH_IN_TYPE, value)

    def till_remove_cash(self, value):
        self._add_voucher(CASH_OUT_TYPE, value)

    def till_read_memory(self, start, end):
        self._send_command(CMD_READ_MEMORY,
                           '%s%sI' % (start.strftime('%d%m%y'),
                                      end.strftime('%d%m%y')))

    def _till_read_memory_to_serial(self, start, end):
        self._send_command(CMD_READ_MEMORY, '%s%sR' % (
            start.strftime('%d%m%y'), end.strftime('%d%m%y')))

        data = ''
        while True:
            c = self.read(1)
            if ord(c) == 0x03:
                break

            data += unicode(c, self.coupon_printer_charset)

        return data

    def till_read_memory_by_reductions(self, start, end):
        self._send_command(CMD_READ_MEMORY,
                           '%06d%06dI' % (start, end))

    def coupon_identify_customer(self, customer, address, document):
        self._customer_name = customer
        self._customer_document = document
        self._customer_address = address

    def coupon_is_customer_identified(self):
        return len(self._customer_document) > 0

    def has_open_coupon(self):
        value = ord(self._read_register(self.registers.FISCAL_FLAGS))
        return value & 1

    def coupon_open(self):
        """ This needs to be called before anything else """
        self._send_command(CMD_COUPON_OPEN,
                           "%-29s%-30s%-80s" % (self._customer_document,
                                                self._customer_name,
                                                self._customer_address))

    def coupon_cancel(self):
        """ Can only be called when a coupon is opened. It needs to be possible
        to open new coupons after this is called.
        """
        self._send_command(CMD_COUPON_CANCEL)

    def cancel_last_coupon(self):
        """Cancel the last non fiscal coupon or the last sale."""
        value = ord(self._read_register(self.registers.FISCAL_FLAGS))
        if value & ALLOW_CANCEL_FISCAL_COUPON:
            self._send_command(CMD_COUPON_CANCEL)
        else:
            self._send_command(CMD_CANCEL_LAST)

    def coupon_close(self, message=""):
        """  This can only be called when the coupon is open, has items added,
        payments added and totalized is called. It needs to be possible to open
        new coupons after this is called.
        """
        self._send_command(CMD_COUPON_CLOSE, message[:CHARS_LIMIT])
        self._reset()
        return self._get_coupon_number()

    def coupon_add_item(self, code, description, price, taxcode,
                        quantity=Decimal("1.0"), unit=UnitType.EMPTY,
                        discount=Decimal("0.0"), markup=Decimal("0.0"),
                        unit_desc=""):
        """ The ECF must be configured to round instead to truncate.
        When truncating, the value may be lower then the one calculated
        by stoq. In this case, the payments added will be higher than the ECF
        expects, and a cents change will be printed.
        """
        if unit == UnitType.CUSTOM:
            unit = unit_desc
        else:
            unit = self._consts.get_value(unit)

        data = ("%02s"     # taxcode
                "%09d"     # value
                "%07d"     # quantity
                "%010d"    # discount
                "%010d"    # markup
                "%022d"    # padding
                "%2s"      # unit
                "%-48s\0"  # code
                "%-200s\0"  # description
                % (taxcode,
                   price * Decimal("1e3"),
                   quantity * Decimal("1e3"),
                   discount * Decimal("1e2"),
                   markup * Decimal("1e2"),
                   0, unit, code, description))
        self._send_command(CMD_ADD_ITEM, data)
        return self._get_last_item_id()

    def coupon_cancel_item(self, item_id=None):
        """ Cancel an item added to coupon; if no item id is specified,
        cancel the last item added. """
        last_item = self._get_last_item_id()
        if item_id is None:
            item_id = last_item
        elif item_id not in xrange(1, last_item + 2):
            raise CancelItemError("There is no such item with ID %r"
                                  % item_id)
        self._send_command(CMD_CANCEL_ITEM, "%04d" % (item_id,))

    def coupon_add_payment(self, payment_method, value, description=u""):
        self._send_command(CMD_ADD_PAYMENT,
                           "%s%014d%s" % (payment_method[:2],
                                          int(value * Decimal('1e2')),
                                          description[:80]))
        self.remainder_value -= value
        if self.remainder_value < 0.0:
            self.remainder_value = Decimal("0.0")
        return self.remainder_value

    def coupon_totalize(self, discount=currency(0), markup=currency(0),
                        taxcode=TaxType.NONE):

        if discount:
            type = 'd'
            value = discount
        elif markup:
            type = 'a'
            value = markup
        else:
            # Just to use the StartClosingCoupon in case of no discount/markup
            # be specified.
            type = 'a'
            value = 0

        self._send_command(CMD_COUPON_TOTALIZE, '%s%014d' % (
            type, int(value * Decimal('1e2'))))

        totalized_value = self._get_coupon_subtotal()
        self.remainder_value = totalized_value
        return totalized_value

    def get_payment_receipt_identifier(self, method):
        # We don't need one.
        return None

    def _get_payment_description(self, method_id):
        constants = self.get_payment_constants()
        for value, name in constants:
            if value == method_id:
                return name

    def gerencial_report_open(self):
        self._send_command(CMD_GERENCIAL_REPORT_PRINT)

    def gerencial_report_print(self, text):
        for line in text.split('\n'):
            self._send_command(CMD_GERENCIAL_REPORT_PRINT, line + '\n')

    def gerencial_report_close(self):
        self._send_command(CMD_GERENCIAL_REPORT_CLOSE)

    def payment_receipt_open(self, identifier, coo, method_id, value):
        method = self._get_payment_description(method_id)
        if not method:
            raise DriverError('Looks like this payment method '
                              'is not configured in the printer')

        value = int(value * 100)
        self._send_command(CMD_PAYMENT_RECEIPT_OPEN,
                           '%-16s%014d%06d' % (method, value, coo))

    def payment_receipt_print(self, text):
        for line in text.split('\n'):
            self._send_command(CMD_PAYMENT_RECEIPT_PRINT, line + '\n')

    def payment_receipt_close(self):
        self._send_command(CMD_PAYMENT_RECEIPT_CLOSE)

    def payment_receipt_print_duplicate(self):
        self._send_command(CMD_PAYMENT_RECEIPT_PRINT_DUPLICATE)

    def get_capabilities(self):
        return dict(
            item_code=Capability(max_len=13),
            item_id=Capability(digits=4),
            items_quantity=Capability(min_size=1, digits=4, decimals=3),
            item_price=Capability(digits=6, decimals=2),
            item_description=Capability(max_len=29),
            payment_value=Capability(digits=12, decimals=2),
            promotional_message=Capability(max_len=320),
            payment_description=Capability(max_len=48),
            customer_name=Capability(max_len=30),
            customer_id=Capability(max_len=28),
            customer_address=Capability(max_len=80),
            add_cash_value=Capability(min_size=0.1, digits=12, decimals=2),
            remove_cash_value=Capability(min_size=0.1, digits=12, decimals=2),
        )

    def get_constants(self):
        return self._consts

    def query_status(self):
        #return '\x02\x05\x00\x1b#(f\x00'
        query = self._create_packet(chr(CMD_READ_REGISTER) +
                                    chr(self.registers.SERIAL))
        return query

    def status_reply_complete(self, reply):
        return len(reply) == 25

    def get_serial(self):
        return self._read_register(self.registers.SERIAL).strip('\x00')

    def get_ccf(self):
        return self._read_register(self.registers.CCF)

    def get_coo(self):
        return self._read_register(self.registers.COO)

    def get_gnf(self):
        return self._read_register(self.registers.GNF)

    def get_crz(self):
        return self._read_register(self.registers.NUMBER_REDUCTIONS_Z)

    def get_tax_constants(self):
        status = self._read_register(self.registers.TOTALIZERS)
        status = struct.unpack('>H', status)[0]

        length, data = self._send_command(CMD_READ_TAXCODES, response='b32s')

        constants = []
        for i in range(16):
            value = bcd2dec(data[i * 2:i * 2 + 2])
            if not value:
                continue

            if 1 << 15 - i & status == 0:
                tax = TaxType.CUSTOM
            else:
                tax = TaxType.SERVICE
            constants.append((tax,
                              '%02d' % (i + 1,),
                              Decimal(value) / 100))

        constants.extend([
            (TaxType.SUBSTITUTION, 'FF', None),
            (TaxType.EXEMPTION, 'II', None),
            (TaxType.NONE, 'NN', None),
        ])

        return constants

    def get_payment_constants(self):
        status = self._read_register(self.registers.PAYMENT_METHODS)[1]
        methods = []
        for i in range(20):
            method = status[i * 16:i * 16 + 16]
            if method != '\x00' * 16:
                methods.append(('%02d' % (i + 1), method.strip()))
        return methods

    def get_sintegra(self):
        opening_date = self._read_register(self.registers.EMISSION_DATE)
        cro = self._read_register(self.registers.CRO)
        # FIXME: This is being fetched before the actual reduction, so the value will be wrong by
        # -1
        crz = self._read_register(self.registers.NUMBER_REDUCTIONS_Z)
        coo = self._get_coupon_number()
        total_cancelations = self._read_register(self.registers.TOTAL_CANCELATIONS)
        total_discount = self._read_register(self.registers.TOTAL_DISCOUNT)

        # Avbr function TACBrECFBematech.GetVendaBruta
        registers = self._send_command(62, 55, response='308s')
        coupon_end = int(bcd2hex(registers)[568:568 + 6])

        grande_total = self._read_register(self.registers.TOTAL)
        grande_total = grande_total / Decimal(100)
        total_bruto = bcd2dec(registers[1:10]) / Decimal(100)

        length, names = self._send_command(CMD_READ_TAXCODES, response='b32s')
        status = self._read_register(self.registers.TOTALIZERS)
        status = struct.unpack('>H', status)[0]
        values = self._send_command(CMD_READ_TOTALIZERS, response='219s')

        taxes = []
        for i in range(length):
            if 1 << 15 - i & status != 0:
                type = 'ISS'
            else:
                type = 'ICMS'

            name = bcd2hex(names[i * 2:i * 2 + 2])
            value = bcd2dec(values[i * 7:i * 7 + 7])
            taxes.append((name, value / Decimal(100), type))

        taxes.append(('CANC', total_cancelations / Decimal(100), 'ICMS'))
        taxes.append(('DESC', total_discount / Decimal(100), 'ICMS'))
        taxes.append(('I', bcd2dec(values[112:119]) / Decimal(100), 'ICMS'))
        taxes.append(('N', bcd2dec(values[119:126]) / Decimal(100), 'ICMS'))
        taxes.append(('F', bcd2dec(values[126:133]) / Decimal(100), 'ICMS'))
        date = bcd2hex(opening_date[:6])

        return Settable(
            opening_date=datetime.date(year=2000 + int(date[4:6]),
                                       month=int(date[2:4]),
                                       day=int(date[:2])),
            serial=self.get_serial(),
            serial_id='%03d' % self._read_register(self.registers.NUMBER_TILL),
            coupon_start=0,
            coupon_end=coupon_end,
            cro=cro,
            crz=crz,
            coo=coo,
            period_total=grande_total - total_bruto,
            total=grande_total,
            taxes=taxes)

    def get_firmware_version(self):
        """Return the firmware version"""
        # REGISTER IS AN INTEGER: 10000 and shoud be formated as 01.00.00
        ret = self._read_register(self.registers.FIRMWARE)
        ret = '%0*d' % (6, ret)
        firmware = "%s:%s:%s" % (ret[0:2], ret[2:4], ret[4:6])
        return firmware


if __name__ == "__main__":
    from stoqdrivers.serialbase import SerialPort
    #port = Serial('/dev/pts/5')
    port = SerialPort('/tmp/com2')
    p = MP25(port)

    p._setup_constants()
